import os


def run_script(script_path, param_path, output_folder):
    """
    Run a python script with the given parameters.

    Parameters
    ----------
    script_path : str
        Path to the python script to run.
    param_path : str
        Path to the parameter file to use.
    output_folder : str
        Path to the folder where the output files will be saved.
    """
    # Check if scripts exist
    if not os.path.exists(script_path):
        raise FileNotFoundError(f"Script '{script_path}' does not exist.")
    if not os.path.exists(param_path):
        raise FileNotFoundError(f"Parameter file '{param_path}' does not exist.")

    # Check if scripts are python scripts
    if not script_path.endswith('.py'):
        raise ValueError(f"Script '{script_path}' is not a python script.")
    if not param_path.endswith('.py'):
        raise ValueError(f"Parameter file '{param_path}' is not a python script.")

    # Run the script
    os.system(f"python {script_path} {param_path} {output_folder}")


class Simulation:
    def __init__(self, param_folder, python_folder, output_folder):

        # Check if param and python folders exist
        if not os.path.exists(param_folder):
            raise FileNotFoundError(f"Parameter folder '{param_folder}' does not exist.")
        if not os.path.exists(python_folder):
            raise FileNotFoundError(f"Python folder '{python_folder}' does not exist.")

        # Check if output folder exists, if not, create it
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)

        self.param_folder = param_folder
        self.python_folder = python_folder
        self.output_folder = output_folder


    def network(self, script_path, param_path):
        run_script(os.path.join(self.python_folder, script_path),
                   os.path.join(self.param_folder, param_path),
                   self.output_folder)

    def simulate(self, script_path, param_path):
        run_script(os.path.join(self.python_folder, script_path),
                   os.path.join(self.param_folder, param_path),
                   self.output_folder)


    def analysis(self, script_path, param_path):
        run_script(os.path.join(self.python_folder, script_path),
                   os.path.join(self.param_folder, param_path),
                   self.output_folder)


import os
import pickle
import sys
import time
import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as ss
from importlib import util
import ncpi
from ncpi import tools

# Choose to either download files and precomputed outputs used in simulations of the reference multicompartment neuron
# network model (True) or load them from a local path (False)
zenodo_dw_mult = True

# Zenodo URL that contains the data (used if zenodo_dw_mult is True)
zenodo_URL_mult = "https://zenodo.org/api/records/15429373"

# Zenodo directory where the data is stored (must be an absolute path to correctly load morphologies in neuron)
zenodo_dir = '/DATA/multicompartment_neuron_network'

# Download data
if zenodo_dw_mult:
    print('\n--- Downloading data.')
    start_time = time.time()
    tools.download_zenodo_record(zenodo_URL_mult, download_dir=zenodo_dir)
    end_time = time.time()
    print(f"All files downloaded in {(end_time - start_time) / 60:.2f} minutes.")


def get_spike_rate(times, transient, dt, tstop):
    """
    Compute the spike rate from spike times.

    Parameters
    ----------
    times : array
        Spike times.
    transient : float
        Transient time at the start of the simulation.
    dt : float
        Simulation time step.
    tstop : float
        Simulation stop time.

    Returns
    -------
    bins : array
        Time bins.
    hist : array
        Spike rate.
    """
    bins = np.arange(transient, tstop + dt, dt)
    hist, _ = np.histogram(times, bins=bins)
    return bins, hist.astype(float)

def get_mean_spike_rate(times, transient, tstop):
    """
    Compute the mean firing rate.

    Parameters
    ----------
    times : array
        Spike times.
    transient : float
        Transient time at the start of the simulation.
    tstop : float
        Simulation stop time.

    Returns
    -------
    float
        Mean firing rate.
    """
    return times.size / (tstop - transient) * 1000

def zscore(data, ch, time):
    """
    Compute the z score using the maximum value of all channels instead of
    the standard deviation of the sample.

    Parameters
    ----------
    data : list
        List of data arrays.
    ch : int
        Channel to normalize.
    time : list
        Time array to normalize.

    Returns
    -------
    tuple
        Maximum value and normalized data.
    """
    tr_data = np.array(data)[:,time]
    tr_data -= np.mean(tr_data,axis = 1).reshape(-1,1)
    return (np.max(np.abs(tr_data)),tr_data[ch] /np.max(np.abs(tr_data)))


if __name__ == "__main__":
    # Read the script file path from sys.argv[1]
    script_path = sys.argv[1]

    # Add the directory containing the script to the Python path
    script_dir = os.path.dirname(script_path)
    sys.path.append(script_dir)

    # Import the script as a module
    module_name = os.path.basename(script_path).replace('.py', '')
    spec = util.spec_from_file_location(module_name, script_path)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Transient time
    transient = module.KernelParams.transient

    # load tstop
    with open(os.path.join(sys.argv[2],'tstop.pkl'), 'rb') as f:
        tstop = pickle.load(f)

    # load dt
    with open(os.path.join(sys.argv[2],'dt.pkl'), 'rb') as f:
        dt = pickle.load(f)

    # Load spike times
    with open(os.path.join(sys.argv[2],'times.pkl'), 'rb') as f:
        times = pickle.load(f)

    # Load gids
    with open(os.path.join(sys.argv[2],'gids.pkl'), 'rb') as f:
        gids = pickle.load(f)

    # Load X and N_X
    with open(os.path.join(sys.argv[2],'network.pkl'), 'rb') as f:
        LIF_params = pickle.load(f)
        P_X = LIF_params['X']
        N_X = LIF_params['N_X']

    # Simulation output from the multicompartment neuron network model
    output_path = os.path.join(zenodo_dir, 'multicompartment_neuron_network', 'output',
                               'adb947bfb931a5a8d09ad078a6d256b0')

    # Path to the data files of the multicompartment neuron models
    multicompartment_neuron_network_path = os.path.join(zenodo_dir, 'multicompartment_neuron_network')

    # Compute the kernel
    print('Computing the kernel...')
    potential = ncpi.FieldPotential()
    biophys = ['set_Ih_linearized_hay2011','make_cell_uniform']
    H_YX = potential.create_kernel(multicompartment_neuron_network_path,
                                   output_path,
                                   module.KernelParams,
                                   biophys,
                                   dt,
                                   tstop,
                                   electrodeParameters=module.KernelParams.electrodeParameters,
                                   CDM=True)

    # Compute LFP signals
    probe = 'GaussCylinderPotential'
    LFP_data = dict(EE=[], EI=[], IE=[], II=[])

    for X in P_X:
        for Y in P_X:
            # Compute the firing rate
            bins, spike_rate = get_spike_rate(times[X], transient, dt, tstop)
            n_ch = H_YX[f'{X}:{Y}'][probe].shape[0]
            for ch in range(n_ch):
                # LFP kernel at electrode 'ch'
                kernel = H_YX[f'{X}:{Y}'][probe][ch, :]
                # LFP signal
                sig = np.convolve(spike_rate, kernel, 'same')
                # Decimate signal (x10)
                LFP_data[f'{X}{Y}'].append(ss.decimate(
                    sig,
                    q=10,
                    zero_phase=True))

    # Compute CDM
    probe = 'KernelApproxCurrentDipoleMoment'
    CDM_data = dict(EE=[], EI=[], IE=[], II=[])

    for X in P_X:
        for Y in P_X:
            # Compute the firing rate
            bins, spike_rate = get_spike_rate(times[X], transient, dt, tstop)
            # Pick only the z-component of the CDM kernel
            kernel = H_YX[f'{X}:{Y}'][probe][2, :]
            # CDM
            sig = np.convolve(spike_rate, kernel, 'same')
            CDM_data[f'{X}{Y}'] = ss.decimate(sig,
                                              q=10,
                                              zero_phase=True)

    # Plot spikes and firing rates
    fig = plt.figure(figsize=[6,5], dpi=300)
    ax1 = fig.add_axes([0.15,0.45,0.75,0.5])
    ax2 = fig.add_axes([0.15,0.08,0.75,0.3])
    # Time interval
    T = [4000, 4100]

    for i, Y in enumerate(P_X):
        #  Compute the mean firing rate
        mean_spike_rate = get_mean_spike_rate(times[Y], transient, tstop)

        t = times[Y]
        gi = gids[Y]
        gi = gi[t >= transient]
        t = t[t >= transient]

        # Spikes
        ii = (t >= T[0]) & (t <= T[1])
        ax1.plot(t[ii], gi[ii], '.',
                 mfc='C{}'.format(i),
                 mec='w',
                 label=r'$\langle \nu_\mathrm{%s} \rangle =%.2f$ s$^{-1}$' % (
                    Y, mean_spike_rate / N_X[i])
                 )
    ax1.legend(loc=1)
    ax1.axis('tight')
    ax1.set_xticklabels([])
    ax1.set_ylabel('gid', labelpad=0)

    # Rates
    for i, Y in enumerate(P_X):
        # Compute the firing rate
        bins, spike_rate = get_spike_rate(times[Y], transient, dt, tstop)
        # Plot the firing rate
        bins = bins[:-1]
        ii = (bins >= T[0]) & (bins <= T[1])
        ax2.plot(bins[ii], spike_rate[ii], color='C{}'.format(i),
                 label=r'$\nu_\mathrm{%s}$' % Y)

    ax2.axis('tight')
    ax2.set_xlabel('t (ms)', labelpad=0)
    ax2.set_ylabel(r'$\nu_X$ (spikes/$\Delta t$)', labelpad=0)

    # Plot kernels and LFP/CDM data
    # Create figure and panels
    fig1 = plt.figure(figsize=[7, 6], dpi=150)
    fig2 = plt.figure(figsize=[7, 6], dpi=150)
    ax1 = []
    ax2 = []

    # First row
    for k in range(4):
        ax1.append(fig1.add_axes([0.1 + k * 0.22, 0.4, 0.18, 0.5]))
        ax2.append(fig2.add_axes([0.1 + k * 0.22, 0.4, 0.18, 0.5]))
    # Second row
    for k in range(4):
        ax1.append(fig1.add_axes([0.1 + k * 0.22, 0.1, 0.18, 0.2]))
        ax2.append(fig2.add_axes([0.1 + k * 0.22, 0.1, 0.18, 0.2]))

    # Time arrays
    dt *= 10  # take into account the decimate ratio
    bins = bins[::10]  # take into account the decimate ratio
    time = np.arange(-module.KernelParams.tau, module.KernelParams.tau + dt, dt)
    T = [4000, 4100]
    ii = (bins >= T[0]) & (bins <= T[1])
    iii = np.where(bins >= T[0] + np.diff(T)[0] / 2)[0][0]

    # LFP probe
    probe = 'GaussCylinderPotential'
    k = 0
    for X in P_X:
        for Y in P_X:
            n_ch = H_YX[f'{X}:{Y}'][probe].shape[0]
            for ch in range(n_ch):
                # Decimate first
                dec_kernel = ss.decimate(H_YX[f'{X}:{Y}'][probe],
                                         q=10, zero_phase=True)
                # Z-scored kernel from 0 to 1/2 of tau
                maxk, norm_ker = zscore(dec_kernel, ch,
                                        np.arange(int(time.shape[0] / 2),
                                                  int(3 * time.shape[0] / 4)))
                # Z-scored LFP signal from T[0] to T[1]
                maxs, norm_sig = zscore(LFP_data[f'{X}{Y}'], ch, ii)
                # Plot data stacked in the Z-axis
                ax1[k].plot(time[int(time.shape[0] / 2):int(3 * time.shape[0] / 4)],
                            norm_ker - ch)
                ax2[k].plot(bins[ii], norm_sig - ch)
            ax1[k].set_title(f'H_{X}:{Y}')
            ax2[k].set_title(f'H_{X}:{Y}')
            if k == 0:
                ax1[k].set_yticks(np.arange(0, -n_ch, -1))
                ax1[k].set_yticklabels(['ch. ' + str(ch) for ch in np.arange(1, n_ch + 1)])
                ax2[k].set_yticks(np.arange(0, -n_ch, -1))
                ax2[k].set_yticklabels(['ch. ' + str(ch) for ch in np.arange(1, n_ch + 1)])
            else:
                ax1[k].set_yticks([])
                ax1[k].set_yticklabels([])
                ax2[k].set_yticks([])
                ax2[k].set_yticklabels([])
            ax1[k].set_xlabel(r'$tau_{ms}$')
            ax2[k].set_xlabel('t (ms)')

            # Add scales
            ax1[k].plot([time[int(0.59 * time.shape[0])], time[int(0.59 * time.shape[0])]],
                        [0, -1], linewidth=2., color='k')
            sexp = np.round(np.log2(maxk))
            ax1[k].text(time[int(0.6 * time.shape[0])], -0.5, r'$2^{%s}mV$' % sexp)
            # ax2[k].plot([bins[iii],bins[iii]],
            #              [0,-1],linewidth = 2., color = 'k')
            # sexp = np.round(np.log2(maxs))
            # ax2[k].text(bins[iii],-0.5,r'$2^{%s}mV$' % sexp)
            k += 1

    # Current dipole moment
    probe = 'KernelApproxCurrentDipoleMoment'
    k = 0
    for X in P_X:
        for Y in P_X:
            # Pick only the z-component of the CDM kernel.
            # (* 1E-4 : nAum --> nAcm unit conversion)
            dec_kernel = ss.decimate([1E-4 * H_YX[f'{X}:{Y}'][probe][2]],
                                     q=10, zero_phase=True)
            maxk, norm_ker = zscore(dec_kernel, 0,
                                    np.arange(int(time.shape[0] / 2),
                                              int(3 * time.shape[0] / 4)))
            # Z-scored CDM signal
            maxs, norm_sig = zscore([1E-4 * CDM_data[f'{X}{Y}']], 0, ii)
            # Plot data
            ax1[k + 4].plot(time[int(time.shape[0] / 2):int(3 * time.shape[0] / 4)], norm_ker)
            ax2[k + 4].plot(bins[ii], norm_sig)

            if k == 0:
                ax1[k + 4].set_yticks([0])
                ax1[k + 4].set_yticklabels([r'$P_z$'])
                ax2[k + 4].set_yticks([0])
                ax2[k + 4].set_yticklabels([r'$P_z$'])
            else:
                ax1[k + 4].set_yticks([])
                ax1[k + 4].set_yticklabels([])
                ax2[k + 4].set_yticks([])
                ax2[k + 4].set_yticklabels([])
            ax1[k + 4].set_xlabel(r'$tau_{ms}$')
            ax2[k + 4].set_xlabel('t (ms)')

            # Add scales
            ax1[k + 4].plot([time[int(0.59 * time.shape[0])], time[int(0.59 * time.shape[0])]],
                            [0, -1], linewidth=2., color='k')
            sexp = np.round(np.log2(maxk))
            ax1[k + 4].text(time[int(0.6 * time.shape[0])], -0.5, r'$2^{%s}nAcm$' % sexp)
            ax2[k + 4].plot([bins[iii + 5], bins[iii + 5]],
                            [0, -1], linewidth=2., color='k')
            sexp = np.round(np.log2(maxs))
            ax2[k + 4].text(bins[iii], -0.5, r'$2^{%s}nAcm$' % sexp)
            k += 1

    plt.show()



import os
import pickle
import sys
from importlib import util

if __name__ == "__main__":
    # Read the script file path from sys.argv[1]
    script_path = sys.argv[1]

    # Add the directory containing the script to the Python path
    script_dir = os.path.dirname(script_path)
    sys.path.append(script_dir)

    # Import the script as a module
    module_name = os.path.basename(script_path).replace('.py', '')
    spec = util.spec_from_file_location(module_name, script_path)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # LIF network parameters
    LIF_params = module.LIF_params

    # Save data to a pickle file
    with open(os.path.join(sys.argv[2],'network.pkl'), 'wb') as f:
        pickle.dump(LIF_params, f)




import os
import pickle
import sys
import nest
import time
import numpy as np
from importlib import util

class LIF_network(object):
    """
    Class to create and simulate a LIF network model.
    """

    def __init__(self, LIF_params):
        # Parameters of the LIF network model
        self.LIF_params = LIF_params

    def create_LIF_network(self, local_num_threads, dt):
        """
        Create network nodes and connections.
        """

        nest.ResetKernel()
        nest.SetKernelStatus(
            dict(
                local_num_threads=local_num_threads,
                rng_seed=int(1+time.localtime().tm_mon *\
                             time.localtime().tm_mday *\
                             time.localtime().tm_hour *\
                             time.localtime().tm_min *\
                             time.localtime().tm_sec *\
                             np.random.rand(1)[0]),
                resolution=dt,
                tics_per_ms=1000 / dt))

        # Neurons
        self.neurons = {}
        for (X, N, C_m, tau_m, E_L, (tau_syn_ex, tau_syn_in)
             ) in zip(self.LIF_params['X'], self.LIF_params['N_X'],
                      self.LIF_params['C_m_X'], self.LIF_params['tau_m_X'],
                      self.LIF_params['E_L_X'], self.LIF_params['tau_syn_YX']):
            net_params = dict(
                C_m=C_m,
                tau_m=tau_m,
                E_L=E_L,
                V_reset=E_L,
                tau_syn_ex=tau_syn_ex,
                tau_syn_in=tau_syn_in
            )
            print('Creating population %s, tau_syn_ex = %s, tau_syn_in = %s\n' % (
                                                        X,tau_syn_ex,tau_syn_in),
                                                        end=' ', flush=True)
            self.neurons[X] = nest.Create(self.LIF_params['model'],
                                          N, net_params)

        # Poisson generators
        self.poisson = {}
        for X, n_ext in zip(self.LIF_params['X'], self.LIF_params['n_ext']):
            self.poisson[X] = nest.Create(
                'poisson_generator', 1, dict(
                    rate=self.LIF_params['nu_ext'] * n_ext))

        # Spike recorders
        self.spike_recorders = {}
        for X in self.LIF_params['X']:
            self.spike_recorders[X] = nest.Create('spike_recorder', 1)

        # Connections
        for i, X in enumerate(self.LIF_params['X']):
            # Recurrent connections
            for j, Y in enumerate(self.LIF_params['X']):
                conn_spec = dict(
                    rule='pairwise_bernoulli',
                    p=self.LIF_params['C_YX'][i][j],
                )
                print('Connecting %s with %s with weight %s\n' % (X,Y,
                                                self.LIF_params['J_YX'][i][j]),
                                                end=' ', flush=True)
                syn_spec = dict(
                    synapse_model='static_synapse',
                    weight=nest.math.redraw(
                        nest.random.normal(
                            mean=self.LIF_params['J_YX'][i][j],
                            std=abs(self.LIF_params['J_YX'][i][j]) * 0.1,
                        ),
                        min=0. if self.LIF_params['J_YX'][i][j] >= 0 else -np.inf,
                        max=np.inf if self.LIF_params['J_YX'][i][j] >= 0 else 0.,
                    ),

                    delay=nest.math.redraw(
                        nest.random.normal(
                            mean=self.LIF_params['delay_YX'][i][j],
                            std=self.LIF_params['delay_YX'][i][j] * 0.5,
                        ),
                        min=0.3,
                        max=np.inf,
                    )
                )

                nest.Connect(
                    self.neurons[X],
                    self.neurons[Y],
                    conn_spec,
                    syn_spec)

            # Poisson generators
            nest.Connect(
                self.poisson[X],
                self.neurons[X],
                'all_to_all',
                dict(
                    weight=self.LIF_params['J_ext']))

            # Recorders
            nest.Connect(self.neurons[X], self.spike_recorders[X])

    def simulate(self, tstop):
        """Instantiate and run simulation"""
        nest.Simulate(tstop)


if __name__ == "__main__":
    # Read the script file path from sys.argv[1]
    script_path = sys.argv[1]

    # Add the directory containing the script to the Python path
    script_dir = os.path.dirname(script_path)
    sys.path.append(script_dir)

    # Import the script as a module
    module_name = os.path.basename(script_path).replace('.py', '')
    spec = util.spec_from_file_location(module_name, script_path)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Simulation time
    tstop = module.tstop

    # Number of threads
    local_num_threads = module.local_num_threads

    # Simulation time step
    dt = module.dt

    # Load network parameters
    with open(os.path.join(sys.argv[2],'network.pkl'), 'rb') as f:
        LIF_params = pickle.load(f)

    # Create the LIF network model
    network = LIF_network(LIF_params)
    network.create_LIF_network(local_num_threads, dt)

    # Simulation
    print('Simulating...\n', end=' ', flush=True)
    tac = time.time()
    network.simulate(tstop)
    toc = time.time()
    print(f'The simulation took {toc - tac} seconds.\n', end=' ', flush=True)

    # Get spike times
    times = dict()
    for i, X in enumerate(network.LIF_params['X']):
        times[X] = nest.GetStatus(network.spike_recorders[X])[0]['events']['times']

    # Get gids
    gids = dict()
    for i, X in enumerate(network.LIF_params['X']):
        gids[X] = nest.GetStatus(network.spike_recorders[X])[0]['events']['senders']

    # Save spike times
    with open(os.path.join(sys.argv[2],'times.pkl'), 'wb') as f:
        pickle.dump(times, f)

    # Save gids
    with open(os.path.join(sys.argv[2],'gids.pkl'), 'wb') as f:
        pickle.dump(gids, f)

    # Save tstop
    with open(os.path.join(sys.argv[2],'tstop.pkl'), 'wb') as f:
        pickle.dump(tstop, f)

    # Save dt
    with open(os.path.join(sys.argv[2],'dt.pkl'), 'wb') as f:
        pickle.dump(dt, f)



import os
import pickle
import numpy as np
import time
import matplotlib.pyplot as plt
import seaborn as sns
from ncpi import tools

# Choose to either download data from Zenodo (True) or load it from a local path (False).
# Important: the zenodo downloads will take a while, so if you have already downloaded the data, set this to False and
# configure the zenodo_dir variable to point to the local path where the data is stored.
zenodo_dw_sim = True # simulation data

# Zenodo URL that contains the simulation data and ML models (used if zenodo_dw_sim is True)
zenodo_URL_sim = "https://zenodo.org/api/records/15351118"

# Paths to zenodo files
zenodo_dir_sim = "zenodo_sim_files"

# Download simulation data and ML models
if zenodo_dw_sim:
    print('\n--- Downloading simulation data and ML models from Zenodo.')
    start_time = time.time()
    tools.download_zenodo_record(zenodo_URL_sim, download_dir=zenodo_dir_sim)
    end_time = time.time()
    print(f"All files downloaded in {(end_time - start_time) / 60:.2f} minutes.")


if __name__ == "__main__":
    # Analyze parameters of the simulation data
    for method in ['catch22']:
        print(f'\n\n--- Method: {method}')

        # Load parameters of the model (theta) and features (X) from simulation data
        print('\n--- Loading simulation data.')
        start_time = time.time()

        try:
            with open(os.path.join(zenodo_dir_sim, 'data', method, 'sim_theta'), 'rb') as file:
                theta = pickle.load(file)
            with open(os.path.join(zenodo_dir_sim, 'data', method, 'sim_X'), 'rb') as file:
                X = pickle.load(file)
        except Exception as e:
            print(f"Error loading simulation data: {e}")

        # Print info
        print('theta:')
        for key, value in theta.items():
            if isinstance(value, np.ndarray):
                print(f'--Shape of {key}: {value.shape}')
            else:
                print(f'--{key}: {value}')
        print(f'Shape of X: {X.shape}')

        # Plot some statistics of the simulation data
        plt.figure(dpi = 300)
        plt.rc('font', size=8)
        plt.rc('font', family='Arial')

        # 1D histograms
        for param in range(7):
            print(f'Parameter {theta["parameters"][param]}')
            plt.subplot(2,4,param+1)
            ax = sns.histplot(theta['data'][:,param], kde=True, bins=50, color='blue')
            ax.set_title(f'Parameter {theta["parameters"][param]}')
            ax.set_xlabel('')
            ax.set_ylabel('')
            plt.tight_layout()

        plt.figure(figsize=(15, 15))
        plt.rc('font', size=8)
        plt.rc('font', family='Arial')

        # 2D histograms
        for i in range(7):
            for j in range(i + 1, 7):
                print(f'Parameter {theta["parameters"][i]} vs Parameter {theta["parameters"][j]}')
                plt.subplot(7, 7, i * 7 + j + 1)
                hist, xedges, yedges = np.histogram2d(theta['data'][:, i], theta['data'][:, j], bins=50)
                plt.imshow(hist.T, origin='lower', interpolation='bilinear', cmap='viridis', aspect='auto',
                           extent=[xedges[0], xedges[-1], yedges[0], yedges[-1]])
                plt.colorbar()
                plt.xlabel(f'{theta["parameters"][i]}')
                plt.ylabel(f'{theta["parameters"][j]}')
                plt.title(f'{theta["parameters"][i]} vs {theta["parameters"][j]}')
                plt.tight_layout()

        plt.show()


import time
import os
import pickle
import numpy as np
import ncpi

# Path to folder where simulation features are stored
sim_file_path = 'zenodo_sim_files/data'

# Choose whether to use a held-out dataset or the full dataset
held_out_dataset = False

# List of parameters to be included in the training
# Full list of parameters:
params =  ['J_EE', 'J_IE', 'J_EI', 'J_II', 'tau_syn_E', 'tau_syn_I', 'J_ext']

# Special case that includes the E/I parameter and the synaptic time constants
# params = ['E_I', 'tau_syn_E', 'tau_syn_I']

# Special case that only includes the E/I parameter
# params = ['E_I']

# ML model to train
model = 'MLPRegressor'

# List of feature sets used to train the model
# all_methods = ['catch22','power_spectrum_parameterization_1', 'SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1',
#                'SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1', 'MD_hrv_classic_pnn40', 'catch22_psp_1']
all_methods = ['catch22', 'power_spectrum_parameterization_1']

# Names of catch22 features
try:
    import pycatch22
    catch22_names = pycatch22.catch22_all([0])['names']
except:
    catch22_names = ['DN_HistogramMode_5',
                     'DN_HistogramMode_10',
                     'CO_f1ecac',
                     'CO_FirstMin_ac',
                     'CO_HistogramAMI_even_2_5',
                     'CO_trev_1_num',
                     'MD_hrv_classic_pnn40',
                     'SB_BinaryStats_mean_longstretch1',
                     'SB_TransitionMatrix_3ac_sumdiagcov',
                     'PD_PeriodicityWang_th0_01',
                     'CO_Embed2_Dist_tau_d_expfit_meandiff',
                     'IN_AutoMutualInfoStats_40_gaussian_fmmi',
                     'FC_LocalSimple_mean1_tauresrat',
                     'DN_OutlierInclude_p_001_mdrmd',
                     'DN_OutlierInclude_n_001_mdrmd',
                     'SP_Summaries_welch_rect_area_5_1',
                     'SB_BinaryStats_diff_longstretch0',
                     'SB_MotifThree_quantile_hh',
                     'SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1',
                     'SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1',
                     'SP_Summaries_welch_rect_centroid',
                     'FC_LocalSimple_mean3_stderr']

if __name__ == "__main__":
    # Iterate over the methods used to compute the features
    for method in all_methods:
        print(f'\n\n--- Method: {method}')
        # Load parameters of the model (theta) and features from simulation data (X)
        print('\n--- Loading simulation data.')
        start_time = time.time()

        # Choose either the catch22 features or the power spectrum parameterization features
        if method != 'catch22_psp_1':
            if method == 'catch22' or method in catch22_names:
                folder = 'catch22'
            else:
                folder = method

            # Load parameters of the model (theta) and features (X) from simulation data
            try:
                with open(os.path.join(sim_file_path, folder, 'sim_theta'), 'rb') as file:
                    theta = pickle.load(file)
                with open(os.path.join(sim_file_path, folder, 'sim_X'), 'rb') as file:
                    X = pickle.load(file)
            except Exception as e:
                print(f"Error loading simulation data: {e}")
                
            if method in catch22_names:
                X = X[:, catch22_names.index(method)]
                print(f'X shape: {X.shape}, column selected: {catch22_names.index(method)}')

        # Concatenate both catch22 and power spectrum parameterization features
        else:
            try:
                with open(os.path.join(sim_file_path, 'catch22', 'sim_theta'), 'rb') as file:
                    theta = pickle.load(file)
                with open(os.path.join(sim_file_path, 'catch22', 'sim_X'), 'rb') as file:
                    X_1 = pickle.load(file)
                with open(os.path.join(sim_file_path, 'power_spectrum_parameterization_1', 'sim_X'), 'rb') as file:
                    X_2 = pickle.load(file)
            except Exception as e:
                print(f"Error loading simulation data: {e}")

            X = np.concatenate((X_1, X_2.reshape(-1,1)), axis=1)
            print(f'X shape: {X.shape}')

        # Subset the data to include only the parameters of interest
        if params is not None:
            print(f'\n--- Subsetting the data to include only the parameters of interest.')
            # First compute the E/I parameter
            if 'E_I' in params:
                E_I = (theta['data'][:,0]/theta['data'][:,2]) / (theta['data'][:,1]/theta['data'][:,3])
                E_I = np.reshape(E_I, (-1, 1))
                theta['data'] = np.concatenate((E_I, theta['data']), axis=1)
                theta['parameters'] = ['E_I'] + theta['parameters']

            # Next generate the subset of parameters
            theta['data'] = theta['data'][:, [theta['parameters'].index(param) for param in params]]
            theta['parameters'] = params
            print(f'Subset of parameters: {params}')

            # If theta['data'] is a 2D array with one column, reshape it to a 1D array
            if theta['data'].shape[1] == 1:
                theta['data'] = np.reshape(theta['data'], (-1,))

        print(f'Shape of theta data: {theta["data"].shape}')
        end_time = time.time()
        print(f'Done in {(end_time - start_time)/60.} min')

        # Create a directory to save results
        if not os.path.exists('data'):
            os.makedirs('data')
        if not os.path.exists(os.path.join('data', method)):
            os.makedirs(os.path.join('data', method))

        # Create a held-out dataset (90% training, 10% testing)
        if held_out_dataset:
            print('\n--- Creating a held-out dataset.')
            np.random.seed(0)
            start_time = time.time()
            indices = np.arange(len(theta['data']))
            np.random.shuffle(indices)
            split = int(0.9 * len(indices))
            train_indices = indices[:split]
            test_indices = indices[split:]

            X_train = X[train_indices]
            X_test = X[test_indices]
            theta_train = theta['data'][train_indices]
            theta_test = theta['data'][test_indices]
            end_time = time.time()

            # Save the held-out dataset
            with open(os.path.join('data', method, 'held_out_dataset'), 'wb') as file:
                pickle.dump((X_test, theta_test), file)
            print(f'\n--- The held-out dataset has been saved.')
            print(f'Done in {(end_time - start_time)/60.} min')

        else:
            print('\n--- Using the full dataset.')
            X_train = X
            X_test = None
            theta_train = theta['data']
            theta_test = None

        # Create the Inference object, add the simulation data and train the model
        print('\n--- Training the regression model.')
        start_time = time.time()

        if model == 'MLPRegressor':
            print('--- Using MLPRegressor')
            if method == 'catch22' or method == 'catch22_psp_1':
                hyperparams = [{'hidden_layer_sizes': (25,25), 'max_iter': 100, 'tol': 1e-1, 'n_iter_no_change': 5},
                               {'hidden_layer_sizes': (50,50), 'max_iter': 100, 'tol': 1e-1, 'n_iter_no_change': 5}]
            else:
                hyperparams = [{'hidden_layer_sizes': (2,2), 'max_iter': 100, 'tol': 1e-1, 'n_iter_no_change': 5},
                               {'hidden_layer_sizes': (4,4), 'max_iter': 100, 'tol': 1e-1, 'n_iter_no_change': 5}]

        if model == 'NPE':
            print('--- Using NPE')
            if method == 'catch22' or method == 'catch22_psp_1':
                hyperparams = [{'prior': None, 'density_estimator': {'model':"maf", 'hidden_features':10,
                                                                     'num_transforms':2}}]
            else:
                hyperparams = [{'prior': None, 'density_estimator': {'model':"maf", 'hidden_features':2,
                                                                     'num_transforms':2}}]

        if model == 'Ridge':
            print('--- Using Ridge')
            hyperparams = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1.}, {'alpha': 10.}, {'alpha': 100.}]


        inference = ncpi.Inference(model=model)
        inference.add_simulation_data(X_train, theta_train)

        # Train the model
        if model == 'NPE':
            # inference.train(param_grid=None, train_params={
            #     'learning_rate': 1e-1,
            #     'stop_after_epochs': 5,
            #     'max_num_epochs': 100})
            # inference.train(param_grid=None)
            inference.train(param_grid=hyperparams, n_splits=10, n_repeats=1)
        else:
            inference.train(param_grid=hyperparams,n_splits=10, n_repeats=20)

        # Save the best model and the StandardScaler
        pickle.dump(pickle.load(open('data/model.pkl', 'rb')),
                    open(os.path.join('data', method, 'model'), 'wb'))
        pickle.dump(pickle.load(open('data/scaler.pkl', 'rb')),
                    open(os.path.join('data', method, 'scaler'), 'wb'))

        # Save density estimator
        if model == 'NPE':
            pickle.dump(pickle.load(open('data/density_estimator.pkl', 'rb')),
                        open(os.path.join('data', method, 'density_estimator'), 'wb'))

        end_time = time.time()
        print(f'Done in {(end_time - start_time)/60.} min')

        # Evaluate the model using the test data
        if held_out_dataset:
            print('\n--- Evaluating the model.')
            start_time = time.time()

            # Predict the parameters from the test data
            predictions = inference.predict(X_test)

            # Save predictions
            with open(os.path.join('data', method, 'predictions'), 'wb') as file:
                pickle.dump(predictions, file)

            end_time = time.time()
            print(f'Done in {(end_time - start_time)/60.} min')


import os
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import matplotlib.lines as mlines
import ncpi

# Set the path to the results folder
results_path = '../data'

# Select the statistical analysis method ('cohen', 'lmer')
statistical_analysis = 'lmer'

# Set the p-value threshold
p_value_th = 0.01

def append_lmer_results(lmer_results, group, elec, p_value_th, data_lmer):
    '''
    Create a list with the z-scores of the linear mixed model analysis for a given group and electrode.

    Parameters
    ----------
    lmer_results : dict
        Dictionary with the results of the linear mixed model analysis.
    group : str
        Group name.
    elec : int
        Electrode index.
    p_value_th : float
        P-value threshold.
    data_lmer : list
        List with the z-scores of the linear mixed model analysis.

    Returns
    -------
    data_lmer : list
        List with the z-scores of the linear mixed model analysis.
    '''

    p_value = lmer_results[f'{group}vsHC']['p.value'].iloc[elec]
    z_score = lmer_results[f'{group}vsHC']['z.ratio'].iloc[elec]

    if p_value < p_value_th:
        data_lmer.append(z_score)
    else:
        data_lmer.append(0)

    return data_lmer


if __name__ == "__main__":
    # Some parameters for the figure
    ncols = 6 
    nrows = 5

    left = 0.06
    right = 0.11

    width = (1.0 - left - right) / (6) - 0.03 
    height = 1.0 / 5 - 0.025
    bottom = 1 - (1. / 5 + 0.07)

    new_spacing_x = 0.08 
    new_spacing_y = 0.05

    spacing_x = 0.04
    spacing_y = 0.064 

    # Create figure
    fig1 = plt.figure(figsize=(7.5, 5.5), dpi=300)
    current_bottom = bottom

    for row in range(2):
        ax = fig1.add_axes([0.01, 0.51 - row * 0.47, 0.98, 0.46 if row == 0 else 0.47])
        ax.add_patch(plt.Rectangle((0, 0), 1, 1, color='red' if row == 0 else 'blue', alpha=0.1))
        ax.set_xticks([])
        ax.set_yticks([])

    for row in range(nrows):
        if row == 0 or row == 1:
            method = 'catch22'
        if row == 2 or row == 3:
            method = 'power_spectrum_parameterization_1'
        try:
            data = pd.read_pickle(os.path.join('../data', method, 'emp_data_reduced.pkl'))
            
        except Exception as e:
            print(f'Error loading data for {method}: {e}')
            continue

        current_left = left
        for col in range(ncols):
            if col == 0 or col == 3:
                group = 'ADMIL'
                group_label = 'ADMIL'

            if col == 1 or col == 4:
                group = 'ADMOD'
                group_label = 'ADMOD'

            if col == 2 or col == 5:
                group = 'ADSEV'
                group_label = 'ADSEV'

            # Add ax --> [left, bottom, width, height]
            ax1 = fig1.add_axes([current_left, current_bottom, width, height], frameon=False)
            
            # Compute new left position (x spacing)
            if col == 2:
                # More spacing for separate plots 
                current_left += width + new_spacing_x
            else:
                current_left += width + spacing_x

            # Disable ticks
            ax1.set_xticks([])
            ax1.set_yticks([])

            # Titles 
            ax1.set_title(f'{group_label} vs HC', fontsize=10)

            # Labels
            if col < 3:
                var = 0 if row == 0 or row == 2 else 1

            if col >= 3:
                var = 3 if row == 0 or row == 2 else 2
    
            # Statistical analysis
            analysis = ncpi.Analysis(data)
            if statistical_analysis == 'lmer':
                stat_results = analysis.lmer(control_group='HC', data_col='Predictions', data_index=var,
                                        other_col=['ID', 'Group', 'Epoch', 'Sensor'],
                                        models={
                                            'mod00': 'Y ~ Group * Sensor + (1 | ID)',
                                            'mod01': 'Y ~ Group * Sensor',
                                            'mod02': 'Y ~ Group + Sensor + (1 | ID)',
                                            'mod03': 'Y ~ Group + Sensor',
                                        },
                                        bic_models=["mod00", "mod01"],
                                        anova_tests={
                                            'test1': ["mod00", "mod01"],
                                            'test2': ["mod02", "mod03"],
                                        },
                                        specs='~Group | Sensor')
            elif statistical_analysis == 'cohend':
                stat_results = analysis.cohend(control_group='HC', data_col='Predictions', data_index=var)
            
            data_stat = []
            # Extract sensor names
            empirical_sensors = data['Sensor'].unique()
            for elec in range(19):
                # Find position of the electrode in the stat results
                pos_results = np.where(stat_results[f'{group}vsHC']['Sensor'] == empirical_sensors[elec])[0]
                
                if len(pos_results) > 0:
                    if statistical_analysis == 'lmer':
                        data_stat = append_lmer_results(stat_results, group, pos_results[0], p_value_th, data_stat)
                    elif statistical_analysis == 'cohend':
                        data_stat.append(stat_results[f'{group}vsHC']['d'][pos_results[0]])
                else:
                    data_stat.append(0)

            # Limits
            if statistical_analysis == 'lmer':
                ylims_stat = [-6., 6.]
            else:
                ylims_stat = [-1., 1.]

            # Create brainplot
            analysis = ncpi.Analysis(data_stat)
            analysis.EEG_topographic_plot(
                        electrode_size = 0.6,
                        ax = ax1,
                        fig=fig1,
                        vmin = ylims_stat[0],
                        vmax = ylims_stat[1],
                        label=False
            )


        # Update "y" spacing
        if row == 1:
            # More spacing for separate plots 
            current_bottom -= height + new_spacing_y
        else:
            current_bottom -= height + spacing_y

    # Text and lines
    fontsize = 12
    fig1.text(0.46, 0.94, 'catch22', color='red', alpha=0.5, fontsize=12, fontstyle='italic')
    fig1.text(0.46, 0.48, '1/f slope', color='blue', alpha=0.5, fontsize=12, fontstyle='italic')

    fig1.text(0.015, 0.94, 'A', fontsize=12, fontweight='bold')
    fig1.text(0.015, 0.48, 'B', fontsize=12, fontweight='bold')

    fig1.text(0.24, 0.94, r'$E/I$', ha='center', fontsize=fontsize)
    fig1.text(0.74, 0.94, r'$J_{syn}^{ext}$ (nA)', ha='center', fontsize=fontsize)

    fig1.text(0.24, 0.7, r'$\tau_{syn}^{exc}$ (ms)', ha='center', fontsize=fontsize)
    fig1.text(0.74, 0.7, r'$\tau_{syn}^{inh}$ (ms)', ha='center', fontsize=fontsize)

    # Parameters for 1/f slope
    fig1.text(0.24, 0.48, r'$E/I$', ha='center', fontsize=fontsize)
    fig1.text(0.74, 0.48, r'$J_{syn}^{ext}$ (nA)', ha='center', fontsize=fontsize)

    fig1.text(0.24, 0.245, r'$\tau_{syn}^{exc}$ (ms)', ha='center', fontsize=fontsize)
    fig1.text(0.74, 0.245, r'$\tau_{syn}^{inh}$ (ms)', ha='center', fontsize=fontsize)

    linepos1 = [0.925, 0.925]
    linepos2 = [0.686, 0.686]

    EI_line_c = mlines.Line2D([0.055, 0.46], linepos1, color='black', linewidth=0.5)
    tauexc_line_c = mlines.Line2D([0.055, 0.46], linepos2, color='black', linewidth=0.5)

    Jext_line_c = mlines.Line2D([0.54, 0.945], linepos1, color='black', linewidth=0.5)
    tauinh_line_c = mlines.Line2D([0.54, 0.945], linepos2, color='black', linewidth=0.5)

    # 1/f slope lines
    linepos1 = [0.467, 0.467]
    linepos2 = [0.23, 0.23]

    EI_line_f = mlines.Line2D([0.055, 0.46], linepos1, color='black', linewidth=0.5)
    tauexc_line_f = mlines.Line2D([0.055, 0.46], linepos2, color='black', linewidth=0.5)

    Jext_line_f = mlines.Line2D([0.54, 0.945], linepos1, color='black', linewidth=0.5)
    tauinh_line_f = mlines.Line2D([0.54, 0.945], linepos2, color='black', linewidth=0.5)

    # Add catch22 lines
    fig1.add_artist(EI_line_c)
    fig1.add_artist(Jext_line_c)
    fig1.add_artist(tauexc_line_c)
    fig1.add_artist(tauinh_line_c)

    # Add 1/f slope lines
    fig1.add_artist(EI_line_f)
    fig1.add_artist(Jext_line_f)
    fig1.add_artist(tauexc_line_f)
    fig1.add_artist(tauinh_line_f)

    fig1.savefig('EEG_predictions.png')

    


import subprocess
import os
from pdf2image import convert_from_path

# Python code to be included in the LaTeX document
code = """
import ncpi

# Build the LIF network model and simulate it
sim = ncpi.Simulation(param_folder='params',
                      python_folder='python',
                      output_folder='output')
sim.network('network.py', 'network_params.py')
sim.simulate('simulation.py', 'simulation_params.py')

# Compute the spatiotemporal kernel
potential = ncpi.FieldPotential(kernel=True)
H_YX = potential.create_kernel(MC_model_folder,
                               MC_output_path,
                               kernelParams,
                               biophys,
                               dt,
                               tstop, 
                               electrodeParameters, 
                               CDM=True)

# Compute the CDMs
probe = 'KernelApproxCurrentDipoleMoment'
kernel = H_YX[f'{X}:{Y}'][probe][2, :] # z-axis
CDMs = np.convolve(LIF_spike_rates, kernel, 'same')

# Obtain features from simulation and empirical data
features = ncpi.Features(method='catch22')
sim_df = features.compute_features(CDMs)
emp_df = features.compute_features(emp_data)

# Train the neural network model using 10-fold CV
hyperparams = [{'hidden_layer_sizes': (25,25)},
               {'hidden_layer_sizes': (50,50)}] 
inference = ncpi.Inference(model='MLPRegressor')
inference.add_simulation_data(sim_df['Features'],
                              theta) # parameters
inference.train(param_grid=hyperparams,
                n_splits=10,
                n_repeats=1) 

# Predict the cortical circuit parameters
predictions = inference.predict(emp_df['Features'])

# Perform the LMER analysis
analysis = ncpi.Analysis()
analysis.lmer(predictions)
"""

# Create a LaTeX document
latex_code = f"""
\\documentclass{{article}}
\\usepackage{{minted}}
\\usepackage{{geometry}}

% Set page dimensions to match the image size
\\geometry{{papersize={{3.5in,6.5in}}, margin=0.1in}}

\\begin{{document}}

\\begin{{minted}}[fontsize=\\footnotesize, breaklines]{{python}}
{code}
\\end{{minted}}

\\end{{document}}
"""

# Write the LaTeX code to a file
with open("code_document.tex", "w") as tex_file:
    tex_file.write(latex_code)

# Compile the LaTeX file to PDF
try:
    subprocess.run(
        ["pdflatex", "-shell-escape", "code_document.tex"],
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    print("PDF generated successfully.")
except subprocess.CalledProcessError as e:
    print(f"Error during PDF generation: {e.stderr.decode()}")

# Convert PDF to High-Resolution Image
pdf_file = "code_document.pdf"
output_image_prefix = "output_image"
try:
    # Convert PDF to images
    images = convert_from_path(pdf_file, dpi=300)
    for i, image in enumerate(images):
        image_path = f"{output_image_prefix}_page_{i + 1}.png"
        image.save(image_path, "PNG")
        print(f"Saved high-resolution image: {image_path}")
except Exception as e:
    print(f"Error during PDF-to-image conversion: {e}")

# Step 4: Cleanup intermediate files (optional)
for ext in [".aux", ".log", ".out", ".tex"]:
    try:
        os.remove(f"code_document{ext}")
    except FileNotFoundError:
        pass


import os
import pickle
import time
import pandas as pd
import scipy
import numpy as np
import shutil
import ncpi
from ncpi import tools

# Select either raw EEG data or source-reconstructed EEG data. This study used the raw EEG data for all analyses.
raw = True
if raw:
    data_path = '/DATA/empirical_datasets/POCTEP_data/CLEAN/SENSORS'
else:
    data_path = '/DATA/empirical_datasets/POCTEP_data/CLEAN/SOURCES/dSPM/DK'

# Choose to either download data from Zenodo (True) or load it from a local path (False).
# Important: the zenodo downloads will take a while, so if you have already downloaded the data, set this to False and
# configure the zenodo_dir variables to point to the local paths where the data is stored.
zenodo_dw_sim = True # simulation data

# Zenodo URL that contains the simulation data and ML models (used if zenodo_dw_sim is True)
zenodo_URL_sim = "https://zenodo.org/api/records/15351118"

# Paths to zenodo simulation files
zenodo_dir_sim = "zenodo_sim_files"

# Methods used to compute the features
all_methods = ['catch22','power_spectrum_parameterization_1']

# ML model used to compute the predictions
ML_model = 'MLPRegressor'


def create_POCTEP_dataframe(data_path):
    '''
    Load the POCTEP dataset and create a DataFrame with the data.

    Parameters
    ----------
    data_path: str
        Path to the directory containing the POCTEP data files.

    Returns
    -------
    df : pandas.DataFrame
        DataFrame containing the POCTEP data.
    '''


    # List files in the directory
    ldir = os.listdir(data_path)

    ID = []
    group = []
    epoch = []
    sensor = []
    EEG = []

    for pt,file in enumerate(ldir):
        print(f'\rProcessing {file} - {pt + 1 }/{len(ldir)}', end="", flush=True)

        # load data
        data = scipy.io.loadmat(data_path + '/' + file)['data']
        signal = data['signal'][0, 0]

        # get sampling frequency
        fs = data['cfg'][0, 0]['fs'][0, 0][0, 0]

        # Electrodes (raw data)/ regions (if source data)
        regions = np.arange(signal.shape[1])

        # get channels
        ch_names = data['cfg'][0, 0]['channels'][0, 0][0]
        ch_names = [ch_names[ll][0] for ll in range(len(ch_names))]

        # 5-second epochs
        epochs = np.arange(0, signal.shape[0], int(fs * 5))

        for i in range(len(epochs) - 1):
            ep = signal[epochs[i]:epochs[i + 1], :]
            # z-score normalization
            ep = (ep - np.mean(ep, axis=0)) / np.std(ep, axis=0)

            # Append data
            for rg in regions:
                ID.append(pt)
                group.append(file.split('_')[0])
                epoch.append(i)
                sensor.append(ch_names[rg])
                EEG.append(ep[:, rg])

    # Create the Pandas DataFrame
    df = pd.DataFrame({'ID': ID,
                       'Group': group,
                       'Epoch': epoch,
                       'Sensor': sensor,
                       'Data': EEG})
    df['Recording'] = 'EEG'
    df['fs'] = fs

    return df

if __name__ == "__main__":
    # Download simulation data and ML models
    if zenodo_dw_sim:
        print('\n--- Downloading simulation data and ML models from Zenodo.')
        start_time = time.time()
        tools.download_zenodo_record(zenodo_URL_sim, download_dir=zenodo_dir_sim)
        end_time = time.time()
        print(f"All files downloaded in {(end_time - start_time) / 60:.2f} minutes.")

    # Go through all methods
    for method in all_methods:
        print(f'\n\n--- Method: {method}')

        # Load parameters of the model (theta) and features (X) from simulation data
        print('\n--- Loading simulation data.')
        start_time = time.time()

        try:
            with open(os.path.join(zenodo_dir_sim, 'data', method, 'sim_theta'), 'rb') as file:
                theta = pickle.load(file)
            with open(os.path.join(zenodo_dir_sim, 'data', method, 'sim_X'), 'rb') as file:
                X = pickle.load(file)
        except Exception as e:
            print(f"Error loading simulation data: {e}")

        # Print info
        print('theta:')
        for key, value in theta.items():
            if isinstance(value, np.ndarray):
                print(f'--Shape of {key}: {value.shape}')
            else:
                print(f'--{key}: {value}')
        print(f'Shape of X: {X.shape}')

        end_time = time.time()
        print(f"Done in {(end_time - start_time):.2f} sec.")

        # Load empirical data and create DataFrame
        df = create_POCTEP_dataframe(data_path=data_path)

        ##########################
        #   FEATURE EXTRACTION   #
        ##########################

        start_time = time.time()

        # Parameters of the feature extraction method
        if method == 'catch22':
            params = None
        elif method == 'power_spectrum_parameterization_1':
            fooof_setup_emp = {'peak_threshold': 1.,
                               'min_peak_height': 0.,
                               'max_n_peaks': 5,
                               'peak_width_limits': (10., 50.)}
            params={'fs': emp_data['fs'][0],
                   'fmin': 5.,
                   'fmax': 45.,
                   'fooof_setup': fooof_setup_emp,
                   'r_squared_th':0.9}
            
        features = ncpi.Features(method=method if method == 'catch22' else 'power_spectrum_parameterization',
                                 params=params)
        emp_data = features.compute_features(df)

        # Keep only the aperiodic exponent (1/f slope)
        if method == 'power_spectrum_parameterization_1':
            emp_data['Features'] = emp_data['Features'].apply(lambda x: x[1])

        end_time = time.time()
        print(f'Done in {(end_time - start_time):.2f} sec')

        #######################################################
        #   PREDICTIONS OF PARAMETERS OF THE NEURAL CIRCUIT   #
        #######################################################

        print('\nComputing predictions...')
        start_time = time.time()

        # Add "Predictions" column to later store the parameters infered
        emp_data['Predictions'] = np.nan

        # List of sensors
        sensor_list = [
            'Fp1','Fp2','F3','F4','C3','C4','P3','P4','O1',
            'O2','F7','F8','T3','T4','T5','T6','Fz','Cz','Pz'
        ]

        # Create folder to save results
        if not os.path.exists('data'):
            os.makedirs('data')
        if not os.path.exists(os.path.join('data', method)):
            os.makedirs(os.path.join('data', method))

        # Create inference object
        inference = ncpi.Inference(model=ML_model)
        # Not sure if this is really needed
        inference.add_simulation_data(X, theta['data'])

        for s, sensor in enumerate(sensor_list):
            print(f'--- Sensor: {sensor}')

            shutil.copy(
                os.path.join(zenodo_dir_sim, 'ML_models/EEG', sensor, method, 'model'),
                os.path.join('data', 'model.pkl')
            )

            shutil.copy(
                os.path.join(zenodo_dir_sim, 'ML_models/EEG', sensor, method, 'scaler'),
                os.path.join('data', 'scaler.pkl')
            )

            sensor_df = emp_data[emp_data['Sensor'].isin([sensor, s])]

            predictions = inference.predict(np.array(sensor_df['Features'].to_list()))

            sensor_df['Predictions'] = [list(pred) for pred in predictions]
            emp_data.update(sensor_df['Predictions'])

        end_time = time.time()
        print(f'Done in {(end_time - start_time):.2f} sec')

        # Save the data including predictions of all parameters
        emp_data.to_pickle(os.path.join('data', method, 'emp_data_all.pkl'))

        # Replace parameters of recurrent synaptic conductances with the ratio (E/I)_net
        E_I_net = emp_data['Predictions'].apply(lambda x: (x[0]/x[2]) / (x[1]/x[3]))
        others = emp_data['Predictions'].apply(lambda x: x[4:])
        emp_data['Predictions'] = (np.concatenate((E_I_net.values.reshape(-1,1),
                                                   np.array(others.tolist())), axis=1)).tolist()

        # Save the data including predictions of (E/I)_net
        emp_data.to_pickle(os.path.join('data', method, 'emp_data_reduced.pkl'))
            


import os
import numpy as np
import matplotlib.pyplot as plt

# Path to the folder with prediction results
pred_results = '../data'

# Names of catch22 features
try:
    import pycatch22
    catch22_names = pycatch22.catch22_all([0])['names']
except:
    catch22_names = ['DN_HistogramMode_5',
                     'DN_HistogramMode_10',
                     'CO_f1ecac',
                     'CO_FirstMin_ac',
                     'CO_HistogramAMI_even_2_5',
                     'CO_trev_1_num',
                     'MD_hrv_classic_pnn40',
                     'SB_BinaryStats_mean_longstretch1',
                     'SB_TransitionMatrix_3ac_sumdiagcov',
                     'PD_PeriodicityWang_th0_01',
                     'CO_Embed2_Dist_tau_d_expfit_meandiff',
                     'IN_AutoMutualInfoStats_40_gaussian_fmmi',
                     'FC_LocalSimple_mean1_tauresrat',
                     'DN_OutlierInclude_p_001_mdrmd',
                     'DN_OutlierInclude_n_001_mdrmd',
                     'SP_Summaries_welch_rect_area_5_1',
                     'SB_BinaryStats_diff_longstretch0',
                     'SB_MotifThree_quantile_hh',
                     'SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1',
                     'SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1',
                     'SP_Summaries_welch_rect_centroid',
                     'FC_LocalSimple_mean3_stderr']

# Dictionaries to store the features and ages
emp = {}
ages = {}

# Iterate over the methods used to compute the features
for method in ['catch22', 'power_spectrum_parameterization_1']:
    print(f'\n\n--- Method: {method}')

    # Load empirical data
    try:
        data_EI = np.load(os.path.join(pred_results, method, 'emp_data_reduced.pkl'), allow_pickle=True)
        ages[method] = np.array(data_EI['Group'].tolist())
        # Pick only ages >= 4
        data_EI = data_EI[data_EI['Group'] >= 4]
        ages[method] = ages[method][ages[method] >= 4]
    except:
        raise RuntimeError(f'Error loading empirical data for {method}. Execution stopped.')

    # Remove nan features from empirical data
    if np.array(data_EI['Features'].tolist()).ndim == 1:
        ii = np.where(~np.isnan(np.array(data_EI['Features'].tolist())))[0]
    else:
        ii = np.where(~np.isnan(np.array(data_EI['Features'].tolist())).any(axis=1))[0]
    data_EI = data_EI.iloc[ii]
    ages[method] = ages[method][ii]

    # Collect features
    if method == 'catch22':
        emp['dfa'] = np.array(data_EI['Features'].apply(
            lambda x: x[catch22_names.index('SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1')]).tolist())
        emp['rs_range'] = np.array(data_EI['Features'].apply(
            lambda x: x[catch22_names.index('SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1')]).tolist())
        emp['high_fluct'] = np.array(data_EI['Features'].apply(
            lambda x: x[catch22_names.index('MD_hrv_classic_pnn40')]).tolist())
    elif method == 'power_spectrum_parameterization_1':
        emp['slope'] = np.array(data_EI['Features'].tolist())

# Create a figure and set its properties
fig = plt.figure(figsize=(4, 3), dpi=300)
plt.rcParams.update({'font.size': 10, 'font.family': 'Arial'})
plt.rc('xtick', labelsize=8)
plt.rc('ytick', labelsize=8)

# Define a colormap for empirical data
cmap = plt.colormaps['viridis']

# Plots
for row in range(2):
    for col in range(2):
        ax = fig.add_axes([0.15 + col * 0.5, 0.67 - row * 0.5, 0.29, 0.29])

        # Get the keys for the parameters and features
        if row == 0 and col == 0:
            feat = 'dfa'
            method = 'catch22'
        elif row == 0 and col == 1:
            feat = 'rs_range'
            method = 'catch22'
        elif row == 1 and col == 0:
            feat = 'high_fluct'
            method = 'catch22'
        else:
            feat = 'slope'
            method = 'power_spectrum_parameterization_1'

        # Show empirical data
        try:
            for i, age in enumerate(np.unique(ages[method])):
                idx = np.where(ages[method] == age)[0]
                data_plot = emp[feat][idx]

                # # Boxplot
                # box = ax.boxplot(data_plot, positions=[age], showfliers=False,
                #                  widths=0.9, patch_artist=True, medianprops=dict(color='red', linewidth=0.8),
                #                  whiskerprops=dict(color='black', linewidth=0.5),
                #                  capprops=dict(color='black', linewidth=0.5),
                #                  boxprops=dict(linewidth=0.5))
                # for patch in box['boxes']:
                #     patch.set_facecolor(cmap(i / len(np.unique(ages[method]))))

                # Clip the data between the 5 % and 95 % quantiles
                q1, q3 = np.percentile(data_plot, [5, 95])
                clipped_data = data_plot[(data_plot >= q1) & (data_plot <= q3)]

                # Violin plot
                violin = ax.violinplot(clipped_data, positions=[age], widths=0.9, showextrema=False)

                for pc in violin['bodies']:
                    pc.set_facecolor(cmap(i / len(np.unique(ages[method]))))
                    pc.set_edgecolor('black')
                    pc.set_alpha(0.8)
                    pc.set_linewidth(0.2)

                # violin['cmedians'].set_linewidth(0.6)
                # violin['cmedians'].set_color('red')

                # Boxplot
                box = ax.boxplot(data_plot, positions=[age], showfliers=False,
                                 widths=0.5, patch_artist=True, medianprops=dict(color='red', linewidth=0.8),
                                 whiskerprops=dict(color='black', linewidth=0.5),
                                 capprops=dict(color='black', linewidth=0.5),
                                 boxprops=dict(linewidth=0.5, facecolor=(0, 0, 0, 0)))

                for patch in box['boxes']:
                    patch.set_linewidth(0.2)
        except:
            pass

        # Set labels
        # X-axis labels
        try:
            ax.set_xticks(np.unique(ages[method])[::2])
            ax.set_xticklabels([f'{str(i)}' for i in np.unique(ages[method])[::2]],fontsize = 8)
        except:
            pass

        ax.set_xlabel('Postnatal days')

        # Y-axis labels
        ax.yaxis.set_label_coords(-0.35, 0.5)
        if row == 0 and col == 0:
            ax.set_ylabel(r'$dfa$')
        elif row == 0 and col == 1:
            ax.set_ylabel(r'$rs\ range$')
        elif row == 1 and col == 0:
            ax.set_ylabel(r'$high\ fluct.$')
        else:
            ax.set_ylabel(r'$1/f$' + ' ' + r'$slope$')

# Save the figure
plt.savefig('emp_features.png', bbox_inches='tight')
# plt.show()


import numpy as np
import pandas as pd
import scipy.interpolate
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import re
from matplotlib.cm import ScalarMappable
from ncpi import tools


def extract_variables(formula):
    tokens = re.split(r'[\s\+\~\|\(\)\*\/\-]+', formula)
    return set(t for t in tokens if t and not t.replace('.', '', 1).isdigit())


class Analysis:
    """ The Analysis class is designed to facilitate statistical analysis and data visualization.

    Parameters
    ----------
    data: (list, np.ndarray, pd.DataFrame)
        Data to be analyzed.
    """
    def __init__(self, data):
        self.data = data


    def lmer(self, control_group = 'HC', data_col = 'Y', other_col = ['ID', 'Group', 'Epoch', 'Sensor'],
             data_index = -1, models = None, bic_models = None, anova_tests = None, specs = None):
        """
        Perform linear mixed-effects model (lmer) or linear model (lm) comparisons using R's `lme4`, `emmeans` and
        `nlme` packages. We assume that, at least, a 'Group' column is present in the dataframe.

        Parameters
        ----------
        control_group: str
            The control group to be used for comparisons.
        data_col: str
            The name of the data column to be analyzed.
        other_col: list
            The names of the other columns to be included in the analysis.
            The default is ['ID', 'Group', 'Epoch', 'Sensor'].
        data_index: int
            The index of the data column to be analyzed. If -1, the entire column is used.
        models: dict
            A dictionary of models to be used for analysis. The keys are model names and the values are model formulas.
            if models is None, the default models are used:
                - mod00: Y ~ Group + (1 | ID)
                - mod01: Y ~ Group
            The best model is selected based on BIC (Bayesian Information Criterion), unless bic_models is None.
        bic_models: list
            A  list of models to be evaluated using BIC. If bic_models is None, the first model is selected. All models
            have to be defined in the models dictionary.
            Example:
                bic_models = ["mod01", "mod02"] # Compare mod01 vs. mod02
        anova_tests: dict
            A dictionary that specifies which models should undergo an ANOVA test after BIC selection. If anova_tests
            is None, no ANOVA tests are performed. Each test must contain two models to be compared. All models have
            to be defined in the models dictionary.
            Example:
                anova_tests = {
                    "test1": ["mod00", "mod02"],  # Compare mod00 vs. mod02
                    "test2": ["mod01", "mod03"]   # Compare mod01 vs. mod03
                }
        specs: string
            The specifications for the emmeans function in R. If specs is None, the default specs are used:
                - ~Group

        Returns
        -------
        results: dict
            A dictionary containing the results of the analysis. The keys are the names of the groups being compared
            and the values are DataFrames containing the results of the analysis.

        """
        # Check if rpy2 is installed
        if not tools.ensure_module("rpy2"):
            raise ImportError("rpy2 is required for lmer but is not installed.")
        pandas2ri = tools.dynamic_import("rpy2.robjects.pandas2ri")
        r = tools.dynamic_import("rpy2.robjects","r")
        ListVector = tools.dynamic_import("rpy2.robjects","ListVector")
        ro = tools.dynamic_import("rpy2","robjects")

        # Activate pandas2ri
        pandas2ri.activate()

        # Import R packages
        ro.r('''
        # Function to check and load packages
        load_packages <- function(packages) {
            for (pkg in packages) {
                if (!require(pkg, character.only = TRUE)) {
                    stop("R package '", pkg, "' is not installed.")
                }
            }
        }

        # Load required packages
        load_packages(c("lme4", "emmeans", "nlme"))
        ''')

        # Check if the data is a pandas DataFrame
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError('The data must be a pandas DataFrame.')

        # Check if the data_col is in the DataFrame
        if data_col not in self.data.columns:
            raise ValueError(f'The data_col "{data_col}" is not in the DataFrame columns.')

        # Check if the other_col columns are in the DataFrame
        for col in other_col:
            if col not in self.data.columns:
                raise ValueError(f'The column "{col}" is not in the DataFrame.')

        # Copy the dataframe
        df = self.data.copy()

        # Remove all columns except data_col and other_col
        df = df[other_col + [data_col]]

        # If data_index is not -1, select the data_index value from the data_col
        if data_index >= 0:
            df[data_col] = df[data_col].apply(lambda x: x[data_index])

        # Check if 'Group' is in the DataFrame
        if 'Group' not in df.columns:
            raise ValueError('The column "Group" is not in the DataFrame.')

        # Filter out control_group from the list of unique groups
        groups = df['Group'].unique()
        groups = [group for group in groups if group != control_group]

        # Create a list with the different group comparisons
        groups_comp = [f'{group}vs{control_group}' for group in groups]

        # Remove rows where the data_col is zero
        df = df[df[data_col] != 0]

        # Rename data_col column to Y
        df.rename(columns={data_col: 'Y'}, inplace=True)

        # Force categorical data type for Sensor
        if 'Sensor' in df.columns:
            df["Sensor"] = df["Sensor"].astype(str).astype('category')

        # Default models if none are provided
        if models is None:
            models = {
                'mod00': 'Y ~ Group + (1 | ID)',
                'mod01': 'Y ~ Group'
            }

        # Default specs if none are provided
        if specs is None:
            specs = '~Group'

        # Check that all models defined in bic_models have been also included in the models dictionary
        if bic_models is not None:
            for model in bic_models:
                if model not in models.keys():
                    raise ValueError(f'bic_models: the model "{model}" is not defined in the models dictionary.')

        # Check that all models defined in anova_tests have been also included in the models dictionary
        if anova_tests is not None:
            for test in anova_tests.values():
                for model in test:
                    if model not in models.keys():
                        raise ValueError(f'anova_tests: the model "{model}" is not defined in the models dictionary.')

        # Check that variables included in models are also in the dataframe
        df_columns = set(df.columns)
        missing_vars = {}
        for name, formula in models.items():
            vars_in_formula = extract_variables(formula)
            missing = vars_in_formula - df_columns
            if missing:
                missing_vars[name] = missing

        if missing_vars:
            raise ValueError(f"Some models have missing variables: {missing_vars}")

        results = {}
        for label, label_comp in zip(groups, groups_comp):
            print(f'\n\n--- Group: {label}')
            r('rm(list = ls())')
            df_pair = df[df['Group'].isin([control_group, label])]
            ro.globalenv['df_pair'] = pandas2ri.py2rpy(df_pair)
            ro.globalenv['label'] = label
            ro.globalenv['control_group'] = control_group

            # # Convert to factors
            # r('''
            # df_pair$ID = as.factor(df_pair$ID)
            # df_pair$Group = factor(df_pair$Group, levels = c(label, control_group))
            # df_pair$Epoch = as.factor(df_pair$Epoch)
            # df_pair$Sensor = as.factor(df_pair$Sensor)
            # print(table(df_pair$Group))
            # ''')

            # Convert to factors
            r_code = []
            for col in other_col:
                if col == 'Group':
                    r_code.append(f'df_pair${col} = factor(df_pair${col}, levels = c("{label}", "{control_group}"))')
                else:
                    r_code.append(f'df_pair${col} = as.factor(df_pair${col})')

            # Print table for 'Group'
            if 'Group' in other_col:
                r_code.append('print(table(df_pair$Group))')

            # Join all lines into one R script string
            full_r_script = '\n'.join(r_code)

            # Pass to R
            r(full_r_script)

            # if table in R is empty for any group, skip the analysis
            if r('table(df_pair$Group)')[0] == 0 or r('table(df_pair$Group)')[1] == 0:
                results[label_comp] = pd.DataFrame({'p.value': [1], 'z.ratio': [0]})
            # Fit the linear (mixed-effects) models
            else:
                for ii, (model_name, formula) in enumerate(models.items()):
                    if (bic_models is None and ii == 0) or (bic_models is not None and model_name in bic_models):
                        # print(f'--- BIC test: fitting model: {model_name}')
                        ro.globalenv[model_name] = formula
                        r(f"{model_name} <- {'lmer' if '(1 | ID)' in formula else 'lm'}({model_name}, data=df_pair)")

                # BIC test: handle single and multiple model cases properly
                r('''
                all_models <- names(which(sapply(ls(), function(x) inherits(get(x), "merMod") || inherits(get(x), "lm"))))

                if (length(all_models) == 1) {
                    m_sel <- all_models[1]  # Use the only model available
                } else {
                    bics <- sapply(all_models, function(m) BIC(get(m)))
                    index <- which.min(bics)
                    m_sel <- all_models[index]
                }
                
                final_model <- get(m_sel)

                ''')
                print(f'--- BIC test. Selected model: {r("m_sel")}')
                # Perform ANOVA tests only for user-specified comparisons
                if anova_tests is not None:
                    # Fit the remaining models
                    for ii, (model_name, formula) in enumerate(models.items()):
                        # Check if model already exists in the R environment
                        if model_name not in r.ls():
                            # print(f'--- ANOVA test: fitting model: {model_name}')
                            ro.globalenv[model_name] = formula
                            r(f"{model_name} <- {'lmer' if '(1 | ID)' in formula else 'lm'}({model_name}, data=df_pair)")

                    # Convert to R list
                    r_anova_tests = ListVector(anova_tests)

                    # Assign to R global environment
                    ro.globalenv['anova_tests'] = r_anova_tests

                    r('''
                    
                    m_name <- m_sel
                    for (comparison in names(anova_tests)) {
                        models <- anova_tests[[comparison]]
                        
                        # Check if the selected model is in the list of models to compare
                        if (m_sel %in% models) {
                            anova_result <- tryCatch(
                            {
                                model1 <- get(as.character(models[2]))
                                model2 <- get(as.character(models[1]))
                                
                                # Print model classes for debugging
                                print(paste("model1 class:", paste(class(model1), collapse = ", ")))
                                print(paste("model2 class:", paste(class(model2), collapse = ", ")))
                                
                                # Check if both models are `lm` or `merMod`
                                if (inherits(model1, "merMod") || inherits(model2, "merMod")) {
                                    # Case 1: One model is `lm`, the other is `lmer` 
                                    if (inherits(model1, "lm") && inherits(model2, "merMod")) {
                                        # Convert `lm` → `gls`
                                        formula1 <- formula(model1)
                                        data1 <- model.frame(model1)
                                        model1 <- gls(formula1, data = data1, method = "ML")
                                        
                                        # Convert `lmer` → `lme` (if possible)
                                        formula2 <- formula(model2)
                                        data2 <- model.frame(model2)
                                        random_effect <- findbars(formula2)  # Extract random effects
                                        
                                        if (length(random_effect) == 1) {
                                            # Simple random intercept (e.g., `(1 | Group)`)
                                            model2 <- lme(
                                                fixed = nobars(formula2),
                                                random = as.formula(paste("~1 |", deparse(random_effect[[1]][[3]]))),
                                                data = data2,
                                                method = "ML"
                                            )
                                        } else {
                                            # If random effects are complex, drop them and refit as `gls`
                                            model2 <- gls(nobars(formula2), data = data2, method = "ML")
                                        }
                                    } 
                                    else if (inherits(model2, "lm") && inherits(model1, "merMod")) {
                                        # Same as above, but swap model1/model2
                                        formula2 <- formula(model2)
                                        data2 <- model.frame(model2)
                                        model2 <- gls(formula2, data = data2, method = "ML")
                                        
                                        formula1 <- formula(model1)
                                        data1 <- model.frame(model1)
                                        random_effect <- findbars(formula1)
                                        
                                        if (length(random_effect) == 1) {
                                            model1 <- lme(
                                                fixed = nobars(formula1),
                                                random = as.formula(paste("~1 |", deparse(random_effect[[1]][[3]]))),
                                                data = data1,
                                                method = "ML"
                                            )
                                        } else {
                                            model1 <- gls(nobars(formula1), data = data1, method = "ML")
                                        }
                                    }
                                    
                                    # Now compare using `nlme::anova.lme`
                                    capture.output(nlme::anova.lme(model1, model2))
                                } else {
                                    # Default: Compare two `lm` models
                                    capture.output(stats::anova(model1, model2))
                                }

                            },
                            error = function(e) {
                                cat("\n--- ANOVA Error ---\n")
                                print(e)
                                NULL
                            }
                            )
                                                        
                            # Extract p-value from the ANOVA result.
                            # Case 1: Standard ANOVA table with Pr(>F)
                            if (any(grepl("Pr\\\\(>F\\\\)", anova_result))) {
                                p_line <- anova_result[grep("Pr\\\\(>F\\\\)", anova_result)+2]
                            }
                            # Case 2: Mixed effects output with Pr(>Chisq)
                            else if (any(grepl("Pr\\\\(>Chisq\\\\)", anova_result))) {
                                p_line <- anova_result[grep("Pr\\\\(>Chisq\\\\)", anova_result) + 2]
                            }
                            # are there any other cases?
                            else {
                            next
                            }
                            matches <- regmatches(p_line, gregexpr("[0-9]+\\\\.?[0-9]*(e[+-]?[0-9]+)?", p_line))[[1]]
                            p_value <- as.numeric(tail(matches, 1))
                             
                            
                            if (!is.na(p_value) && p_value >= 0.05) {
                                # Determine which model is simpler (counts all fixed-effect terms, including interactions)
                                formula1 <- formula(get(as.character(models[1])))
                                formula2 <- formula(get(as.character(models[2])))
                                
                                # Count terms (excluding random effects after '|')
                                terms1 <- length(attr(terms(formula1), "term.labels"))  # Handles *, :, etc.
                                terms2 <- length(attr(terms(formula2), "term.labels"))

                                # Select the model with less complexity
                                if (terms1 <= terms2) {
                                    final_model <- get(as.character(models[1]))
                                    m_name <- models[1]
                                } else {
                                    final_model <- get(as.character(models[2]))
                                    m_name <- models[2]
                                }
                                print(class(final_model))
                                break
                            }
                        }
                    }
                    ''')
                    print(f'--- ANOVA test. Selected model: {r("m_name")}')

                # Compute pairwise comparisons
                ro.globalenv['specs'] = specs

                r('''
                emm <- suppressMessages(emmeans(final_model, specs=as.formula(specs)))
                res <- pairs(emm, adjust='holm')
                df_res <- as.data.frame(res)
                ''')

                # Ensure Sensor remains as a character column
                if 'Sensor' in r('names(df_res)'):
                    r('''
                        df_res$Sensor <- as.character(df_res$Sensor)
                    ''')

                df_res_r = ro.r['df_res']
                with (pandas2ri.converter + pandas2ri.converter).context():
                    df_res_pd = pandas2ri.conversion.get_conversion().rpy2py(df_res_r)

                results[label_comp] = df_res_pd

        return results

    def cohend(self, control_group='HC', data_col='Y', data_index=-1):
        '''
        Compute Cohen's d for all pairwise group comparisons across sensors.

        Parameters
        ----------
        control_group: str
            The control group to be used for comparisons.
        data_col: str
            The name of the data column to be analyzed.
        data_index: int
            The index of the data column to be analyzed. If -1, the entire column is used.

        Returns
        -------
        results: dict
            A dictionary containing the results of the analysis. The keys are the names of the groups being compared
            and the values are lists containing the Cohen's d values for each sensor.
        '''

        # Check if the data is a pandas DataFrame
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError('The data must be a pandas DataFrame.')

        # Check if the data_col is in the DataFrame
        if data_col not in self.data.columns:
            raise ValueError(f'The data_col "{data_col}" is not in the DataFrame columns.')

        # Check if 'Group' and 'Sensor' are in the DataFrame
        for col in ['Group', 'Sensor']:
            if col not in self.data.columns:
                raise ValueError(f'The column "{col}" is not in the DataFrame.')

        # Copy the dataframe
        df = self.data.copy()

        # Remove all columns except 'Group', 'Sensor' and data_col
        df = df[['Group', 'Sensor', data_col]]

        # If data_index is not -1, select the data_index value from the data_col
        if data_index >= 0:
            df[data_col] = df[data_col].apply(lambda x: x[data_index])

        # Filter out control_group from the list of unique groups
        groups = df['Group'].unique()
        groups = [group for group in groups if group != control_group]

        # Create a list with the different group comparisons
        groups_comp = [f'{group}vs{control_group}' for group in groups]

        # Remove rows where the data_col is zero
        df = df[df[data_col] != 0]

        results = {}
        for label, label_comp in zip(groups, groups_comp):
            print(f'\n\n--- Group: {label}')

            # filter out control_group and the current group
            df_pair = df[df['Group'].isin([control_group, label])]

            all_d = []
            all_sensors = []
            for sensor in df_pair['Sensor'].unique():
                df_sensor = df_pair[df_pair['Sensor'] == sensor]

                group1 = np.array(df_sensor[df_sensor['Group'] == label][data_col])
                group2 = np.array(df_sensor[df_sensor['Group'] == control_group][data_col])

                # Check if both groups have more than 2 elements
                if len(group1) > 2 and len(group2) > 2:
                    # Calculate Cohen's d
                    n1, n2 = len(group1), len(group2)
                    mean1, mean2 = np.mean(group1), np.mean(group2)
                    std1, std2 = np.std(group1, ddof=1), np.std(group2, ddof=1)

                    pooled_std = np.sqrt(((n1 - 1) * std1 ** 2 + (n2 - 1) * std2 ** 2) / (n1 + n2 - 2))
                    d = (mean1 - mean2) / pooled_std
                    all_d.append(d)
                else:
                    all_d.append(np.nan)
                all_sensors.append(sensor)

            results[label_comp] = pd.DataFrame({'d': all_d, 'Sensor': all_sensors})

        return results

    def EEG_topographic_plot(self, **kwargs):
        '''
        Generate a topographical plot of EEG data using the 10-20 electrode placement system,
        visualizing activity from 19 or 20 electrodes.

        Parameters
        ----------
        **kwargs: keyword arguments:
            - radius: (float)
                Radius of the head circumference.
            - pos: (float)
                Position of the head on the x-axis.
            - electrode_size: (float)
                Size of the electrodes.
            - label: (bool)
                Show the colorbar label.
            - ax: (matplotlib Axes object)
                Axes object to plot the data.
            - fig: (matplotlib Figure object)
                Figure object to plot the data.
            - vmin: (float)
                Min value used for plotting.
            - vmax: (float)
                Max value used for plotting.
        '''

        # Check if mpl_toolkits is installed
        if not tools.ensure_module("mpl_toolkits"):
            raise ImportError("mpl_toolkits is required for EEG_topographic_plot but is not installed.")
        make_axes_locatable = tools.dynamic_import("mpl_toolkits.axes_grid1",
                                                   "make_axes_locatable")

        default_parameters = {
            'radius': 0.6,
            'pos': 0.0,
            'electrode_size': 0.9,
            'label': True,
            'ax': None,
            'fig': None,
            'vmin': None,
            'vmax': None
        }

        for key in kwargs.keys():
            if key not in default_parameters.keys():
                raise ValueError(f'Invalid parameter: {key}')

        radius = kwargs.get('radius', default_parameters['radius'])
        pos = kwargs.get('pos', default_parameters['pos'])
        electrode_size = kwargs.get('electrode_size', default_parameters['electrode_size'])
        label = kwargs.get('label', default_parameters['label'])
        ax = kwargs.get('ax', default_parameters['ax'])
        fig = kwargs.get('fig', default_parameters['fig'])
        vmin = kwargs.get('vmin', default_parameters['vmin'])
        vmax = kwargs.get('vmax', default_parameters['vmax'])

        if not isinstance(radius, float):
            raise ValueError('The radius parameter must be a float.')
        if not isinstance(pos, float):
            raise ValueError('The pos parameter must be a float.')
        if not isinstance(electrode_size, float):
            raise ValueError('The electrode_size parameter must be a float.')
        if not isinstance(label, bool):
            raise ValueError('The label parameter must be a boolean.')
        if not isinstance(ax, plt.Axes):
            raise ValueError('The ax parameter must be a matplotlib Axes object.')
        if not isinstance(fig, plt.Figure):
            raise ValueError('The fig parameter must be a matplotlib Figure object.')
        if not isinstance(vmin, float):
            raise ValueError('The vmin parameter must be a float.')
        if not isinstance(vmax, float):
            raise ValueError('The vmax parameter must be a float.')
        if not isinstance(self.data, (list, np.ndarray)):
            raise ValueError('The data parameter must be a list or numpy array.')
        if len(self.data) not in [19, 20]:
            raise ValueError('The data parameter must contain 19 or 20 elements.')
        
              
        def plot_simple_head(ax, radius=0.6, pos=0):
            '''
            Plot a simple head model with ears and nose.

            Parameters
            ----------
            ax: matplotlib Axes object
            radius: float,
                radius of the head circumference.
            pos: float
                Position of the head on the x-axis.
            '''

            # Adjust the aspect ratio of the plot
            ax.set_aspect('equal')

            # Head
            head_circle = mpatches.Circle((pos, 0), radius+0.02, edgecolor='k', facecolor='none', linewidth=0.5)
            ax.add_patch(head_circle)

            # Ears
            right_ear = mpatches.FancyBboxPatch([pos + radius + radius / 20, -radius / 10],
                                                radius / 50, radius / 5,
                                                boxstyle=mpatches.BoxStyle("Round", pad=radius / 20),
                                                linewidth=0.5)
            ax.add_patch(right_ear)

            left_ear = mpatches.FancyBboxPatch([pos - radius - radius / 20 - radius / 50, -radius / 10],
                                            radius / 50, radius / 5,
                                            boxstyle=mpatches.BoxStyle("Round", pad=radius / 20),
                                            linewidth=0.5)
            ax.add_patch(left_ear)

            # Nose
            ax.plot([pos - radius / 10, pos, pos + radius / 10], 
                    [radius + 0.02, radius + radius / 10 + 0.02,0.02 + radius], 
                    'k', linewidth=0.5)


        def plot_EEG(data, radius, pos, electrode_size, label, ax, fig, vmin, vmax):
            '''
            Plot the EEG data on the head model as a topographic map.

            Parameters
            ----------
            data: list or np.ndarray of size (19,) or (20,)
                EEG data.
            radius: float
                Radius of the head circumference.
            pos: float
                Position of the head on the x-axis.
            electrode_size: float
                Size of the electrodes.
            label: bool
                Show the colorbar label.
            ax: matplotlib Axes object
                Axes object to plot the data.
            fig: matplotlib Figure object
                Figure object to plot the data.
            vmin: float
                Min value used for plotting.
            vmax: float
                Max value used for plotting.
            '''

            # Check data type
            if not isinstance(data, (list, np.ndarray)):
                raise ValueError('The data must be a list or numpy array.')

            # Check data length
            if len(data) not in [19, 20]:
                raise ValueError('The data must contain 19 or 20 elements.')

            # Coordinates of the EEG electrodes
            koord_dict = {
                'Fp1': [pos - 0.25 * radius, 0.8 * radius],
                'Fp2': [pos + 0.25 * radius, 0.8 * radius],
                'F3': [pos - 0.3 * radius, 0.35 * radius],
                'F4': [pos + 0.3 * radius, 0.35 * radius],
                'C3': [pos - 0.35 * radius, 0.0],
                'C4': [pos + 0.35 * radius, 0.0],
                'P3': [pos - 0.3 * radius, -0.4 * radius],
                'P4': [pos + 0.3 * radius, -0.4 * radius],
                'O1': [pos - 0.35 * radius, -0.8 * radius],
                'O2': [pos + 0.35 * radius, -0.8 * radius],
                'F7': [pos - 0.6 * radius, 0.45 * radius],
                'F8': [pos + 0.6 * radius, 0.45 * radius],
                'T3': [pos - 0.8 * radius, 0.0],
                'T4': [pos + 0.8 * radius, 0.0],
                'T5': [pos - 0.6 * radius, -0.2],
                'T6': [pos + 0.6 * radius, -0.2],
                'Fz': [pos, 0.35 * radius],
                'Cz': [pos, 0.0],
                'Pz': [pos, -0.4 * radius],
                'Oz': [pos, -0.8 * radius]
            }
            
            if len(data) == 19:
                del koord_dict['Oz']
            koord = list(koord_dict.values())

            # Number of points used for interpolation
            N = 100

            # External fake electrodes used for interpolation
            for xx in np.linspace(pos-radius,pos+radius,50):
                koord.append([xx,np.sqrt(radius**2 - (xx)**2)])
                koord.append([xx,-np.sqrt(radius**2 - (xx)**2)])
                data.append(0)
                data.append(0)

            # Interpolate data points
            x,y = [],[]
            for i in koord:
                x.append(i[0])
                y.append(i[1])
            z = data

            xi = np.linspace(-radius, radius, N)
            yi = np.linspace(-radius, radius, N)
            zi = scipy.interpolate.griddata((np.array(x), np.array(y)), z,
                                            (xi[None,:], yi[:,None]), method='cubic')


            # Use different number of levels for the fill and the lines
            CS = ax.contourf(xi, yi, zi, 30, cmap = plt.cm.bwr, zorder = 1,
                             vmin = vmin, vmax = vmax)
            ax.contour(xi, yi, zi, 5, colors ="grey", zorder = 2, linewidths = 0.4,
                       vmin = vmin, vmax = vmax)

            # Make a color bar
            # cbar = fig.colorbar(CS, ax=Vax)
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)

            if np.sum(np.abs(data)) > 2: 
                colorbar = fig.colorbar(ScalarMappable(norm=CS.norm, cmap=CS.cmap), cax=cax)
                colorbar.ax.tick_params(labelsize=8)
                if label == True:
                    colorbar.ax.xaxis.set_label_position('bottom')
                    # bbox = colorbar.ax.get_position()
                    # print(bbox)
                    colorbar.set_label('z-ratio', size=5, labelpad=-15, rotation=0, y=0.)
                    
            else:
                # Hide the colorbar if the data is not significant
                cax.axis('off')

            # Add the EEG electrode positions
            ax.scatter(x[:len(koord_dict)], y[:len(koord_dict)], marker ='o', c ='k', s = electrode_size, zorder = 3)


        plot_simple_head(ax, radius, pos)
        plot_EEG(self.data, radius, pos, electrode_size, label, ax, fig, vmin, vmax)


"""
Initialization file for the ncpi package.
"""

from .Inference import *
from .Features import *
from .Simulation import *
from .FieldPotential import *
from .Analysis import *
from . import tools



import os
import pickle
import numpy as np
import random
from ncpi import tools


class Inference:
    """
    General-purpose class for inferring parameters from simulated or observed features using either
    Bayesian inference (with SBI) or regression (with sklearn).

    Attributes
    ----------
    model : list
        Name and backend library of the chosen model (e.g., ['NPE', 'sbi'] or ['RandomForestRegressor', 'sklearn']).
    hyperparams : dict
        Dictionary of hyperparameters passed to the selected model.
    features : np.ndarray
        Input feature array used for training.
    theta : np.ndarray
        Parameter array (target values to infer).

    Methods
    -------
    __init__(model, hyperparams=None)
        Initializes the class with a model and hyperparameters.
    add_simulation_data(features, parameters)
        Adds training data (features and target parameters).
    initialize_sbi(hyperparams)
        Prepares an SBI inference method (only for sbi-based models).
    train(param_grid=None, n_splits=10, n_repeats=10, train_params={...})
        Trains the model using either sbi or sklearn depending on configuration.
    predict(features)
        Predicts parameters for new input features.
    sample_posterior(x, num_samples=10000)
        Samples from the posterior (only for sbi-based models).
    """

    def __init__(self, model, hyperparams=None):
        """
        Initializes the Inference class with the specified model and hyperparameters.

        Parameters
        ----------
        model : str
            Name of the machine-learning model to use. It can be any of the regression models from sklearn or NPE, NLE,
            NRE from SBI.
        hyperparams : dict, optional
            Dictionary of hyperparameters of the model. The default is None.
        """

        # Ensure that sklearn is installed
        if not tools.ensure_module("sklearn"):
            raise ImportError('sklearn is not installed. Please install it to use the Inference class.')
        self.RepeatedKFold = tools.dynamic_import("sklearn.model_selection", "RepeatedKFold")
        self.StandardScaler = tools.dynamic_import("sklearn.preprocessing", "StandardScaler")
        self.all_estimators = tools.dynamic_import("sklearn.utils", "all_estimators")
        self.RegressorMixin = tools.dynamic_import("sklearn.base", "RegressorMixin")

        # Ensure that sbi and torch are installed
        if model in ['NPE', 'NLE', 'NRE']:
            if not tools.ensure_module("sbi"):
                raise ImportError('sbi is not installed.')
            if not tools.ensure_module("torch"):
                raise ImportError('torch is not installed.')

            # Dynamic imports for SBI components
            self.NPE = tools.dynamic_import("sbi.inference", "NPE")
            self.NLE = tools.dynamic_import("sbi.inference", "NLE")
            self.NRE = tools.dynamic_import("sbi.inference", "NRE")
            self.posterior_nn = tools.dynamic_import("sbi.neural_nets", "posterior_nn")
            self.likelihood_nn = tools.dynamic_import("sbi.neural_nets", "likelihood_nn")
            self.classifier_nn = tools.dynamic_import("sbi.neural_nets", "classifier_nn")
            self.BoxUniform = tools.dynamic_import("sbi.utils", "BoxUniform")
            self.torch = tools.dynamic_import("torch")
            self.model = [model, 'sbi']

        # Check if pathos is installed. If not, use the default Python multiprocessing library
        if not tools.ensure_module("pathos"):
            self.pathos_inst = False
            self.multiprocessing = tools.dynamic_import("multiprocessing")
        else:
            self.pathos_inst = True
            self.pathos = tools.dynamic_import("pathos", "pools")

        # Check if tqdm is installed
        if not tools.ensure_module("tqdm"):
            self.tqdm_inst = False
        else:
            self.tqdm_inst = True
            self.tqdm = tools.dynamic_import("tqdm", "tqdm")

        # Assert that model is a string
        if type(model) is not str:
            raise ValueError('Model must be a string.')

        # Check if model is in the list of regression models from sklearn, or it is one of the SBI methods
        regressors = [estimator for estimator in self.all_estimators() if issubclass(estimator[1], self.RegressorMixin)]
        sbi_models = ['NPE', 'NLE', 'NRE']
        if model not in [regressor[0] for regressor in regressors] + sbi_models:
            raise ValueError(f'{model} not in the list of machine-learning models from sklearn or SBI (NPE, NLE, NRE).')

        # Set model and library
        self.model = [model, 'sbi'] if model in sbi_models else [model, 'sklearn']

        # Check if hyperparameters is a dictionary
        if hyperparams is not None:
            if type(hyperparams) is not dict:
                raise ValueError('Hyperparameters must be a dictionary.')
            # Set hyperparameters
            self.hyperparams = hyperparams
        else:
            self.hyperparams = None

        # Initialize features and parameters
        self.features = []
        self.theta = []

        # Set the number of threads used by PyTorch for SBI models
        if model in ['NPE', 'NLE', 'NRE']:
            torch_threads = int(os.cpu_count()/2)
            self.torch.set_num_threads(torch_threads)


    def add_simulation_data(self, features, parameters):
        """
        Method to add features and parameters to the training data.

        Parameters
        ----------
        features : np.ndarray
            Features.
        parameters : np.ndarray
            Parameters to infer.
        """

        # Assert that features and parameters are numpy arrays
        if type(features) is not np.ndarray:
            raise ValueError('X must be a numpy array.')
        if type(parameters) is not np.ndarray:
            raise ValueError('Y must be a numpy array.')

        # Assert that features and parameters have the same number of rows
        if features.shape[0] != parameters.shape[0]:
            raise ValueError('Features and parameters must have the same number of rows.')

        # Create a mask to identify rows without NaN or Inf values
        if features.ndim == 1 and parameters.ndim == 1:
            mask = np.isfinite(features) & np.isfinite(parameters)
        elif features.ndim == 1:
            mask = np.isfinite(features) & np.all(np.isfinite(parameters), axis=1)
        elif parameters.ndim == 1:
            mask = np.all(np.isfinite(features), axis=1) & np.isfinite(parameters)
        else:
            mask = np.all(np.isfinite(features), axis=1) & np.all(np.isfinite(parameters), axis=1)

        # Apply the mask to filter out rows with NaN or Inf values
        features = features[mask]
        parameters = parameters[mask]

        # Stack features and parameters
        features = np.stack(features)
        parameters = np.stack(parameters)

        # Reshape features if your data has a single feature
        if features.ndim == 1:
            features = features.reshape(-1, 1)

        # Add features and parameters to training data
        self.features = features
        self.theta = parameters

    def initialize_sbi(self, hyperparams):
        """
        Initializes the SBI inference method (NPE, NLE, or NRE) using the appropriate neural estimator.

        Parameters
        ----------
        hyperparams : dict
            Dictionary of hyperparameters required to set up the inference method.
            Must include:
                - 'prior': the prior distribution over parameters (BoxUniform or similar)
                - 'density_estimator': a dictionary containing:
                    - 'model': the neural network type (e.g., 'nsf', 'maf', etc.)
                    - 'hidden_features': number of hidden units per layer
                    - 'num_transforms': number of normalizing flow transformations (only used for NPE and NLE)

        Returns
        -------
        inference : NPE, NLE, or NRE object
            A configured SBI inference object ready for appending simulations and training.
        """
        inference_type = self.model[0].lower()

        if 'density_estimator' not in hyperparams:
            raise ValueError('Missing density_estimator.')
        if 'hidden_features' not in hyperparams['density_estimator']:
            raise ValueError('Missing hidden_features.')
        if 'num_transforms' not in hyperparams['density_estimator'] and inference_type in ['npe', 'nle']:
            raise ValueError('Missing num_transforms.')
        if 'model' not in hyperparams['density_estimator']:
            raise ValueError('Missing model.')
        if 'prior' not in hyperparams:
            raise ValueError('Missing prior.')

        est = hyperparams['density_estimator']
        model = est['model']
        hidden = est['hidden_features']
        if inference_type in ['npe', 'nle']:
            transforms = est.get('num_transforms', 5)

        if inference_type == 'npe':
            estimator_fn = self.posterior_nn(model=model, hidden_features=hidden, num_transforms=transforms)
            inference = self.NPE(prior=hyperparams['prior'], density_estimator=estimator_fn)
        elif inference_type == 'nle':
            estimator_fn = self.likelihood_nn(model=model, hidden_features=hidden, num_transforms=transforms)
            inference = self.NLE(prior=hyperparams['prior'], density_estimator=estimator_fn)
        elif inference_type == 'nre':
            estimator_fn = self.classifier_nn(model=model, hidden_features=hidden)
            inference = self.NRE(prior=hyperparams['prior'], ratio_estimator=estimator_fn)
        else:
            raise ValueError(f"'{self.model[0]} is not a valid SBI model. Choose from NPE, NLE, or NRE.")

        return inference



    def train(self, param_grid=None, n_splits=10, n_repeats=10, train_params={'learning_rate': 0.0005,
                                                                              'training_batch_size': 256}):
        """
        Method to train the model.

        Parameters
        ----------
        param_grid : list of dictionaries, optional
            List of dictionaries of hyperparameters to search over. The default
            is None (no hyperparameter search).
        n_splits : int, optional
            Number of splits for RepeatedKFold cross-validation. The default is 10.
        n_repeats : int, optional
            Number of repeats for RepeatedKFold cross-validation. The default is 10.
        train_params : dict, optional
            Dictionary of training parameters for SBI.
        """

        # Import the sklearn model
        if self.model[1] == 'sklearn':
            regressors = [estimator for estimator in self.all_estimators() if issubclass(estimator[1], self.RegressorMixin)]
            pos = np.where(np.array([regressor[0] for regressor in regressors]) == self.model[0])[0][0]
            cl = str(regressors[pos][1]).split('.')[1]
            exec(f'from sklearn.{cl} import {self.model[0]}')

        # Initialize model with default hyperparameters
        if self.hyperparams is None:
            if self.model[1] == 'sklearn':
                model = eval(f'{self.model[0]}')()
            elif self.model[1] == 'sbi':
                model = self.initialize_sbi({'prior': None, 'density_estimator':  {'model': "maf", 'hidden_features': 10,
                                                                                    'num_transforms': 2}})

        # Initialize model with user-defined hyperparameters
        else:
            if self.model[1] == 'sklearn':
                model = eval(f'{self.model[0]}')(**self.hyperparams)
            elif self.model[1] == 'sbi':
                model = self.initialize_sbi(self.hyperparams)

        # Check if features and parameters are not empty
        if len(self.features) == 0:
            raise ValueError('No features provided.')
        if len(self.theta) == 0:
            raise ValueError('No parameters provided.')

        # Initialize the StandardScaler
        scaler = self.StandardScaler()

        # Fit the StandardScaler
        scaler.fit(self.features)

        # Transform the features
        self.features = scaler.transform(self.features)

        # Remove Nan and Inf values from features
        if self.features.ndim == 1:
            mask = np.isfinite(self.features)
        else:
            mask = np.all(np.isfinite(self.features), axis=1)
        self.features = self.features[mask]
        self.theta = self.theta[mask]

        # Search for the best hyperparameters using RepeatedKFold cross-validation and grid search if param_grid is
        # provided
        if param_grid is not None:
            # Assert that param_grid is a list
            if type(param_grid) is not list:
                raise ValueError('param_grid must be a list.')

            # Loop over each set of hyperparameters
            best_score = np.inf
            best_config = None
            best_fits = None
            for params in param_grid:
                print(f'\n\n--> Hyperparameters: {params}')

                # Initialize RepeatedKFold (added random_state for reproducibility)
                rkf = self.RepeatedKFold(n_splits=n_splits, n_repeats=n_repeats, random_state=0)

                # Loop over each repeat and fold
                mean_scores = []
                fits = []
                for repeat_idx, (train_index, test_index) in enumerate(rkf.split(self.features)):
                    # Print info of repeat and fold
                    print('\n') if self.model[1] == 'sbi' else None
                    print(f'\rRepeat {repeat_idx // n_splits + 1}, Fold {repeat_idx % n_splits + 1}', end='', flush=True)
                    print('\n') if self.model[1] == 'sbi' else None

                    # Split the data
                    X_train, X_test = self.features[train_index], self.features[test_index]
                    Y_train, Y_test = self.theta[train_index], self.theta[test_index]

                    if self.model[1] == 'sklearn':
                        # Set the random state for reproducibility
                        params['random_state'] = repeat_idx // n_splits
                        # Update parameters
                        model.set_params(**params)

                        # Fit the model
                        model.fit(X_train, Y_train)
                        fits.append(model)

                        # Predict the parameters
                        Y_pred = model.predict(X_test)

                        # Compute the mean squared error
                        mse = np.mean((Y_pred - Y_test) ** 2)

                        # Append the mean squared error
                        mean_scores.append(mse)

                    if self.model[1] == 'sbi':
                        # Set the seeds for reproducibility
                        self.torch.manual_seed(repeat_idx)
                        random.seed(repeat_idx)

                        # Re-initialize the SBI object with the new configuration
                        model = self.initialize_sbi(params)

                        # Ensure theta is a 2D array
                        if Y_train.ndim == 1:
                            Y_train = Y_train.reshape(-1, 1)

                        # Append simulations
                        model.append_simulations(
                            self.torch.from_numpy(Y_train.astype(np.float32)),
                            self.torch.from_numpy(X_train.astype(np.float32))
                        )

                        # Train the neural density estimator
                        density_estimator = model.train(**train_params)
                        fits.append([model, density_estimator])

                        # Build the posterior
                        posterior = model.build_posterior(density_estimator)

                        # Loop over all test samples
                        for i in range(len(X_test)):
                            # Sample the posterior
                            x_o = self.torch.from_numpy(np.array(X_test[i], dtype=np.float32).reshape(1, -1))
                            posterior_samples = posterior.sample((5000,), x=x_o, show_progress_bars=False)
                            pred = np.mean(posterior_samples.numpy(), axis=0)
                            # Compute the mean squared error
                            mse = np.mean((pred[0] - Y_test[i]) ** 2)
                            # Append the mean squared error
                            mean_scores.append(mse)

                # Compute the mean of the mean squared errors
                if np.mean(mean_scores) < best_score:
                    best_score = np.mean(mean_scores)
                    best_config = params
                    best_fits = fits

            # Update the model with the best hyperparameters
            if best_config is not None:
                if self.model[1] == 'sklearn':
                    model = best_fits
                if self.model[1] == 'sbi':
                    model = [best_fits[i][0] for i in range(len(best_fits))]
                    density_estimator = [best_fits[i][1] for i in range(len(best_fits))]
                print(f'\n\n--> Best hyperparameters: {best_config}\n')
            else:
                raise ValueError('\nNo best hyperparameters found.\n')

        # Fit the model using all the data
        else:
            if self.model[1] == 'sklearn':
                model.fit(self.features, self.theta)

            if self.model[1] == 'sbi':
                # Ensure theta is a 2D array
                if self.theta.ndim == 1:
                    self.theta = self.theta.reshape(-1, 1)

                # Append simulations
                model.append_simulations(
                    self.torch.from_numpy(self.theta.astype(np.float32)),
                    self.torch.from_numpy(self.features.astype(np.float32))
                )

                # Extract training parameters
                learning_rate = train_params.get("learning_rate", 0.0005)
                training_batch_size = train_params.get("training_batch_size", 256)

                # Train the neural density estimator
                density_estimator = model.train(learning_rate=learning_rate, training_batch_size=training_batch_size)

        # Save the best model and the StandardScaler
        if not os.path.exists('data'):
            os.makedirs('data')
        with open('data/model.pkl', 'wb') as file:
            pickle.dump(model, file)
        with open('data/scaler.pkl', 'wb') as file:
            pickle.dump(scaler, file)

        # Save also the density estimator if the model is SBI
        if self.model[1] == 'sbi':
            with open('data/density_estimator.pkl', 'wb') as file:
                pickle.dump(density_estimator, file)

    def predict(self, features):
        """
        Method to predict the parameters.

        Parameters
        ----------
        features : np.ndarray
            Features.

        Returns
        -------
        predictions : list
            List of predictions.
        """

        def process_batch(batch):
            """
            Function to compute predictions from a batch of features.

            Parameters
            ----------
            batch: tuple
                Tuple containing the batch of features, the StandardScaler and the model (and the posterior if the model
                is SBI).

            Returns
            -------
            predictions: list
                List of predictions
            """
            if self.model[1] == 'sbi':
                batch_index, feat_batch, scaler, model, posterior = batch
            else:
                batch_index, feat_batch, scaler, model = batch
            predictions = []
            for feat in feat_batch:
                # Transform the features
                feat = scaler.transform(feat.reshape(1, -1))

                # Check that feat has no NaN or Inf values
                if np.all(np.isfinite(feat)):
                    # Predict the parameters
                    if self.model[1] == 'sklearn':
                        if type(model) is list:
                            pred = np.mean([m.predict(feat) for m in model], axis=0)
                        else:
                            pred = model.predict(feat)
                        predictions.append(pred[0])

                    if self.model[1] == 'sbi':
                        # Sample the posterior
                        x_o = self.torch.from_numpy(np.array(feat, dtype=np.float32))
                        if self.hyperparams is not None:
                            num_samples = self.hyperparams.get("num_samples", 5000)
                        else:
                            num_samples = 5000

                        if type(posterior) is list:
                            posterior_samples = [post.sample((num_samples,), x=x_o, show_progress_bars=False) for post in posterior]
                            # Compute the mean of the posterior samples
                            pred = np.mean([np.mean(post.numpy(), axis=0) for post in posterior_samples], axis=0)
                        else:
                            posterior_samples = posterior.sample((num_samples,), x=x_o, show_progress_bars=False)
                            # Compute the mean of the posterior samples
                            pred = np.mean(posterior_samples.numpy(), axis=0)
                        predictions.append(pred)

                else:
                    predictions.append([np.nan for _ in range(self.theta.shape[1])])

            # Return the predictions
            return batch_index, predictions

        # Assert that the model has been trained
        if not os.path.exists('data/model.pkl'):
            raise ValueError('Model has not been trained.')

        # Load the best model and the StandardScaler
        with open('data/model.pkl', 'rb') as file:
            model = pickle.load(file)
        with open('data/scaler.pkl', 'rb') as file:
            scaler = pickle.load(file)

        if self.model[1] == 'sbi':
            with open('data/density_estimator.pkl', 'rb') as file:
                density_estimator = pickle.load(file)
                # Build the posterior
                if type(density_estimator) is list:
                    posterior = [model[i].build_posterior(density_estimator[i]) for i in range(len(density_estimator))]
                else:
                    posterior = model.build_posterior(density_estimator)

        # Assert that features is a numpy array
        if type(features) is not np.ndarray:
            raise ValueError('features must be a numpy array.')

        # Stack features
        features = np.stack(features)

        # Split the data into batches using the number of available CPUs
        num_cpus = os.cpu_count()
        if self.model[1] == 'sbi':
            batch_size = len(features)  # to avoid memory issues
        else:
            batch_size = len(features) // num_cpus
        if batch_size == 0:
            batch_size = 1
        batches = [(i, features[i:i + batch_size]) for i in range(0, len(features), batch_size)]

        # Choose the appropriate parallel processing library
        pool_class = self.pathos.ProcessPool if self.pathos_inst else self.multiprocessing.Pool

        # Prepare batch arguments based on model type
        use_posterior = self.model[1] == 'sbi'
        batch_args = [(ii, batch, scaler, model, posterior) if use_posterior else (ii, batch, scaler, model) for
                      ii, batch in batches]

        # Compute predictions in parallel if model is not SBI
        if self.model[1] == 'sbi':
            results = [process_batch(batch_arg) for batch_arg in batch_args]
        else:
            with pool_class(num_cpus) as pool:
                imap_results = pool.imap(process_batch, batch_args)
                results = list(
                    self.tqdm(imap_results, total=len(batches), desc="Computing predictions")) if self.tqdm_inst else list(
                    imap_results)

        # Sort the predictions based on the original index
        results.sort(key=lambda x: x[0])
        predictions = [pred for _, batch_preds in results for pred in batch_preds]

        return predictions

    def sample_posterior(self, x, num_samples=10000):
        """
        Sample from the posterior distribution for a given observation.

        Parameters
        ----------
        x : np.ndarray
            Observed feature vector (1D array).
        num_samples : int, optional
            Number of posterior samples to draw. Default is 10000.

        Returns
        -------
        np.ndarray
            Array of posterior samples.
        """
        if not os.path.exists('data/model.pkl'):
            raise ValueError('Model not trained.')

        with open('data/model.pkl', 'rb') as f:
            model = pickle.load(f)
        with open('data/scaler.pkl', 'rb') as f:
            scaler = pickle.load(f)
        with open('data/density_estimator.pkl', 'rb') as f:
            density_estimator = pickle.load(f)

        if type(density_estimator) is list:
            posterior = [model[i].build_posterior(density_estimator[i]) for i in range(len(density_estimator))]
        else:
            posterior = model.build_posterior(density_estimator)

        x = scaler.transform(x.reshape(1, -1))
        x_tensor = self.torch.from_numpy(x.astype(np.float32))

        if isinstance(posterior, list):
            samples = [p.sample((num_samples,), x=x_tensor).numpy() for p in posterior]
            return np.vstack(samples)
        else:
            samples = posterior.sample((num_samples,), x=x_tensor)
            return samples.numpy()


import ncpi

if __name__ == "__main__":
    # Create a Simulation object
    sim = ncpi.Simulation(param_folder='params', python_folder='python', output_folder='output')

    # Run the network, simulation and analysis scripts
    sim.network('network.py', 'network_params.py')
    sim.simulate('simulation.py', 'simulation_params.py')
    sim.analysis('analysis.py', 'analysis_params.py')



import os
import numpy as np
import pandas as pd
import statsmodels.api as sm
import time
import pickle
from statsmodels.formula.api import ols
from matplotlib import pyplot as plt
from ncpi import tools

# Choose to either download data from Zenodo (True) or load it from a local path (False).
# Important: the zenodo downloads will take a while, so if you have already downloaded the data, set this to False and
# configure the zenodo_dir variable to point to the local path where the data is stored.
zenodo_dw_sim = True # simulation data

# Zenodo URL that contains the simulation data and ML models (used if zenodo_dw_sim is True)
zenodo_URL_sim = "https://zenodo.org/api/records/15351118"

# Paths to zenodo files
zenodo_dir_sim = "zenodo_sim_files"

# Download simulation data and ML models
if zenodo_dw_sim:
    print('\n--- Downloading simulation data and ML models from Zenodo.')
    start_time = time.time()
    tools.download_zenodo_record(zenodo_URL_sim, download_dir=zenodo_dir_sim)
    end_time = time.time()
    print(f"All files downloaded in {(end_time - start_time) / 60:.2f} minutes.")

# Names of catch22 features
try:
    import pycatch22
    catch22_names = pycatch22.catch22_all([0])['names']
except:
    catch22_names = ['DN_HistogramMode_5',
                     'DN_HistogramMode_10',
                     'CO_f1ecac',
                     'CO_FirstMin_ac',
                     'CO_HistogramAMI_even_2_5',
                     'CO_trev_1_num',
                     'MD_hrv_classic_pnn40',
                     'SB_BinaryStats_mean_longstretch1',
                     'SB_TransitionMatrix_3ac_sumdiagcov',
                     'PD_PeriodicityWang_th0_01',
                     'CO_Embed2_Dist_tau_d_expfit_meandiff',
                     'IN_AutoMutualInfoStats_40_gaussian_fmmi',
                     'FC_LocalSimple_mean1_tauresrat',
                     'DN_OutlierInclude_p_001_mdrmd',
                     'DN_OutlierInclude_n_001_mdrmd',
                     'SP_Summaries_welch_rect_area_5_1',
                     'SB_BinaryStats_diff_longstretch0',
                     'SB_MotifThree_quantile_hh',
                     'SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1',
                     'SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1',
                     'SP_Summaries_welch_rect_centroid',
                     'FC_LocalSimple_mean3_stderr']


# Dictionaries to store the features and parameters
features = {}
parameters = {'catch22':{},
              'power_spectrum_parameterization_1':{}}

# Iterate over the methods used to compute the features
for method in ['catch22', 'power_spectrum_parameterization_1']:
    print(f'\n\n--- Method: {method}')

    # Load parameters of the model (theta) and features (X) from simulation data
    print('\n--- Loading simulation data.')
    start_time = time.time()

    try:
        with open(os.path.join(zenodo_dir_sim, 'data', method, 'sim_theta'), 'rb') as file:
            theta = pickle.load(file)
        with open(os.path.join(zenodo_dir_sim, 'data', method, 'sim_X'), 'rb') as file:
            X = pickle.load(file)
    except Exception as e:
        print(f"Error loading simulation data: {e}")

    # Print info
    print('theta:')
    for key, value in theta.items():
        if isinstance(value, np.ndarray):
            print(f'--Shape of {key}: {value.shape}')
        else:
            print(f'--{key}: {value}')
    print(f'Shape of X: {X.shape}')

    # Remove nan features from simulation data
    if X.ndim == 1:
        X = X.reshape(-1, 1)
    ii = np.where(~np.isnan(X).any(axis=1))[0]
    X = X[ii]
    theta['data'] = theta['data'][ii]
    print(f'Number of samples after removing nan features: {len(X)}')

    # Collect parameters
    parameters[method]['E_I'] = ((theta['data'][:, 0] / theta['data'][:, 2]) /
                                 (theta['data'][:, 1] / theta['data'][:, 3]))
    parameters[method]['tau_syn_exc'] = theta['data'][:, 4]
    parameters[method]['tau_syn_inh'] = theta['data'][:, 5]
    parameters[method]['J_syn_ext'] = theta['data'][:, 6]

    # Collect features
    if method == 'catch22':
        features['dfa'] = X[:, catch22_names.index('SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1')]
        features['rs_range'] = X[:, catch22_names.index('SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1')]
        features['high_fluct'] = X[:, catch22_names.index('MD_hrv_classic_pnn40')]

    elif method == 'power_spectrum_parameterization_1':
        features['slope'] = X[:,0]

# Create a figure and set its properties
fig = plt.figure(figsize=(7.5, 4.5), dpi=300)
plt.rcParams.update({'font.size': 10, 'font.family': 'Arial'})
plt.rc('xtick', labelsize=8)
plt.rc('ytick', labelsize=8)

# Labels for the parameters
param_labels = [r'$E/I$', r'$\tau_{syn}^{exc}$ (ms)', r'$\tau_{syn}^{inh}$ (ms)',
          r'$J_{syn}^{ext}$ (nA)']

# Define 4 colormaps for simulation data
cmaps = ['Blues', 'Greens', 'Reds', 'Purples']

# Define a colormap for empirical data
# cmap = plt.colormaps['viridis']

# Plots
for row in range(4):
    for col in range(4):
        ax = fig.add_axes([0.09 + col * 0.23, 0.76 - row * 0.21, 0.18, 0.19])

        # Get the keys for the parameters and features
        if row == 0:
            feat = 'dfa'
            method = 'catch22'
        elif row == 1:
            feat = 'rs_range'
            method = 'catch22'
        elif row == 2:
            feat = 'high_fluct'
            method = 'catch22'
        else:
            feat = 'slope'
            method = 'power_spectrum_parameterization_1'

        if col == 0:
            param = 'E_I'
        elif col == 1:
            param = 'tau_syn_exc'
        elif col == 2:
            param = 'tau_syn_inh'
        elif col == 3:
            param = 'J_syn_ext'

        # Simulation data
        if col < 4:
            try:
                # Constraints for the bins
                if col == 0:
                    minp = 0.1
                    maxp = 12.0
                elif col == 1:
                    minp = 0.1
                    maxp = 2.0
                elif col == 2:
                    minp = 1.
                    maxp = 8.0
                elif col == 3:
                    minp = 10.
                    maxp = 40.
                ii = np.where((parameters[method][param] >= minp) & (parameters[method][param] <= maxp))[0]
                n_bins = np.unique(parameters[method][param][ii]).shape[0]
                if n_bins > 15:
                    n_bins = 15
                bins = np.linspace(minp, maxp, n_bins)

                bin_features = []
                group =  []
                for jj in range(len(bins) - 1):
                    pos = np.where((parameters[method][param] >= bins[jj]) & (parameters[method][param] < bins[jj + 1]))[0]
                    bin_features.extend([features[feat][pos]])
                    group.extend([jj for _ in range(len(pos))])

                    # Clip the data between the 5 % and 95 % quantiles
                    q1, q3 = np.percentile(features[feat][pos], [5, 95])
                    clipped_data = features[feat][pos][(features[feat][pos] >= q1) & (features[feat][pos] <= q3)]

                    # Violin plot
                    violin = ax.violinplot(clipped_data, positions=[jj], widths=0.9, showextrema=False)

                    for pc in violin['bodies']:
                        pc.set_facecolor(plt.get_cmap(cmaps[row])(jj / (len(bins) - 1)))
                        pc.set_edgecolor('black')
                        pc.set_alpha(0.8)
                        pc.set_linewidth(0.2)

                    # violin['cmedians'].set_linewidth(0.6)
                    # violin['cmedians'].set_color('red')

                    # Boxplot
                    box = ax.boxplot(features[feat][pos], positions=[jj], showfliers=False,
                                     widths=0.5, patch_artist=True, medianprops=dict(color='red', linewidth=0.8),
                                     whiskerprops=dict(color='black', linewidth=0.5),
                                     capprops=dict(color='black', linewidth=0.5),
                                     boxprops=dict(linewidth=0.5, facecolor=(0, 0, 0, 0)))

                    for patch in box['boxes']:
                        patch.set_linewidth(0.2)

                # Eta squared
                print(f'\n--- Eta squared for {feat} and {param}.')
                # Create the dataframe
                df = pd.DataFrame({'Feature': np.concatenate(bin_features),
                                   'Group': np.array(group)})

                # Perform ANOVA
                model = ols('Feature ~ C(Group)', data=df).fit()
                anova_table = sm.stats.anova_lm(model, typ=2)

                # Compute eta squared
                ss_between = anova_table['sum_sq']['C(Group)']  # Sum of squares for the groups
                ss_total = anova_table['sum_sq'].sum()  # Total sum of squares
                eta_squared = ss_between / ss_total
                print(f"Eta squared: {eta_squared}")

                # Plot eta squared
                y_max = ax.get_ylim()[1]
                y_min = ax.get_ylim()[0]
                delta = (y_max - y_min) * 0.1
                ax.text((len(np.unique(group))-1)/2., y_max + delta/4.,
                        f'$\eta^2$ = {eta_squared:.3f}' if eta_squared > 0.001 else f'$\eta^2$ < 0.001',
                        ha='center', fontsize=8, color = 'black')

                # Change y-lim
                ax.set_ylim([y_min, y_max + 2 * delta])

            except:
                pass

        # Labels
        if col < 4 and row == 3:
            if col == 0:
                step = 4
            elif col == 1 or col == 2:
                step = 3
            else:
                step = 2

            ax.set_xticks(np.arange(0,len(bins) - 1,step))
            ax.set_xticklabels(['%.2f' % ((bins[jj] + bins[jj + 1]) / 2) for jj in np.arange(0,len(bins) - 1,step)],
                               fontsize = 8)
            ax.set_xlabel(param_labels[col])

        else:
            ax.set_xticks([])
            ax.set_xticklabels([])

        if col == 0:
            ax.yaxis.set_label_coords(-0.35, 0.5)
            if row == 0:
                ax.set_ylabel(r'$dfa$')
            elif row == 1:
                ax.set_ylabel(r'$rs\ range$')
            elif row == 2:
                ax.set_ylabel(r'$high\ fluct.$')
            else:
                ax.set_ylabel(r'$1/f$' + ' ' + r'$slope$')

# Save the figure
plt.savefig('sim_features.png', bbox_inches='tight')
# plt.show()


import os
import pickle
import sys
import numpy as np
import torch
import time
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from scipy.ndimage import gaussian_filter1d
from sklearn.model_selection import RepeatedKFold
from ncpi import tools

# Note: The SBI models were trained using sbi version 0.22.0 with the SNPE algorithm. As of sbi version 0.23.0, the
# required import is no longer supported. To run this code, please downgrade sbi to version 0.22.0.

# Choose whether to compute posteriors and diagnostic metrics (True) or load them from file (False)
compute_metrics = True

# Path to the local directory where the metrics and posteriors will be saved
result_folder = 'SBI_results'

# Choose whether to use a held-out dataset or folds from RepeatedKFold
use_held_out_data = True

# Number of random samples to draw from the posteriors
n_samples = 25000

# Choose to either download data from Zenodo (True) or load it from a local path (False).
# Important: the zenodo downloads will take a while, so if you have already downloaded the data, set this to False and
# configure the zenodo_dir variable to point to the local path where the data is stored.
zenodo_dw_sim = True # simulation data

# Zenodo URL that contains the simulation data and ML models (used if zenodo_dw_sim is True)
zenodo_URL_sim = "https://zenodo.org/api/records/15351118"

# Paths to zenodo files
zenodo_dir_sim = "/DATA/zenodo_sim_files"

# Download simulation data and ML models
if zenodo_dw_sim:
    print('\n--- Downloading simulation data and ML models from Zenodo.')
    start_time = time.time()
    tools.download_zenodo_record(zenodo_URL_sim, download_dir=zenodo_dir_sim)
    end_time = time.time()
    print(f"All files downloaded in {(end_time - start_time) / 60:.2f} minutes.")

def hellinger_distance(p, q):
    """
    Compute the Hellinger Distance between two discrete probability distributions P and Q.

    Parameters:
    p (np.array): Probability distribution P.
    q (np.array): Probability distribution Q.

    Returns:
    float: Hellinger Distance between P and Q.
    """
    # Ensure the inputs are numpy arrays
    p = np.asarray(p)
    q = np.asarray(q)

    # Normalize the distributions to ensure they sum to 1
    p = p / np.sum(p)
    q = q / np.sum(q)

    # Compute the Hellinger Distance
    return np.sqrt(np.sum((np.sqrt(p) - np.sqrt(q)) ** 2)) / np.sqrt(2)

def create_white_to_color_cmap(color):
    """
    Create a colormap that transitions from white to the specified color. Alpha (transparency) is higher for white.

    Parameters
    ----------
    color : str
        Color to transition to. Must be a valid matplotlib color string.

    Returns
    -------
    cmap : LinearSegmentedColormap
        Colormap that transitions from white to the specified color.

    """
    # Define the colormap from white to the specified color
    cmap = LinearSegmentedColormap.from_list('white_to_color', ['white', color])

    # Initialize the colormap's lookup table
    cmap._init()

    # Set alpha (transparency) gradient
    # Use the size of the lookup table (cmap._lut.shape[0]) to ensure compatibility
    cmap._lut[:, -1] = np.linspace(0, 0.6, cmap._lut.shape[0])

    return cmap


# List of methods
all_methods = ['catch22', 'catch22_psp_1', 'SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1',
               'SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1', 'MD_hrv_classic_pnn40',
               'power_spectrum_parameterization_1']

# Set the seeds
np.random.seed(0)
torch.manual_seed(0)

# Path to ML models trained based on a held-out dataset approach
if use_held_out_data:
    ML_path = os.path.join(zenodo_dir_sim, 'ML_models/held_out_data_models')
# Path to ML models trained based on a RepeatedKFold approach
else:
    ML_path = os.path.join(zenodo_dir_sim, 'ML_models/4_param')

# Limits of histograms
lims = [[-15, 15], [-2, 5], [-2, 12], [0, 60]]

if compute_metrics:
    # Dictionaries to store posteriors and diagnostic metrics
    all_post_samples = {}
    all_theta = {}
    z_score = {}
    shrinkage = {}
    abs_error = {}
    PRE = {}

    for method in all_methods:
        print(f'\n\n--- Method: {method}\n')

        # Initialize dictionaries
        all_post_samples[method] = []
        all_theta[method] = []
        z_score[method] = []
        shrinkage[method] = []
        abs_error[method] = []
        PRE[method] = []

        # Load density estimators and inference models
        try:
        # if method == 'catch22' or method == 'catch22_psp_1' or method == 'SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1':
            # Load X and theta from the held-out dataset
            if use_held_out_data:
                print('\n--- Loading the held-out dataset')
                with open(os.path.join(ML_path, 'datasets', method, 'held_out_dataset'), 'rb') as file:
                    X, theta = pickle.load(file)
            # Load X and theta from all folds of RepeatedKFold and concatenate them
            else:
                X = pickle.load(open(os.path.join(zenodo_dir_sim, 'data',method,'sim_X'),'rb'))
                theta = pickle.load(open(os.path.join(zenodo_dir_sim, 'data',method,'sim_theta'),'rb'))
                rkf = RepeatedKFold(n_splits=10, n_repeats=1, random_state=0)
                new_X = []
                new_theta = []
                for repeat_idx, (train_index, test_index) in enumerate(rkf.split(X)):
                    if repeat_idx == 0:
                        new_X = X[test_index]
                        new_theta = theta['data'][test_index]
                    else:
                        new_X = np.concatenate((new_X,X[test_index]),axis=0)
                        new_theta = np.concatenate((new_theta,theta['data'][test_index]),axis=0)
                X = new_X
                theta = new_theta

            # Calculate the E/I ratio
            theta_EI = np.zeros((theta.shape[0], 4))
            theta_EI[:, 0] = (theta[:, 0] / theta[:, 2]) / (theta[:, 1] / theta[:, 3])
            theta_EI[:, 1] = theta[:, 4]
            theta_EI[:, 2] = theta[:, 5]
            theta_EI[:, 3] = theta[:, 6]

            # Variance of the prior (due to the Law of Large Numbers, this variance will be very close to the
            # population variance)
            var_theta = np.var(theta_EI, axis=0)

            print('\n--- Loading SBI models')
            density_estimator = pickle.load(open(os.path.join(ML_path, 'SBI', method, 'density_estimator'), 'rb'))
            model = pickle.load(open(os.path.join(ML_path, 'SBI', method, 'model'), 'rb'))
            scaler = pickle.load(open(os.path.join(ML_path, 'SBI', method, 'scaler'), 'rb'))

            # Print info about the density estimator
            flow = density_estimator[0]  # Extract the Flow model
            transform = flow._transform  # Get the transform object

            # Extract the first MaskedAffineAutoregressiveTransform
            maf_transform = transform._transforms[1]  # First MAF layer

            # Extract the autoregressive network (MADE)
            made = maf_transform.autoregressive_net

            # Input features
            input_features = made.context_layer.in_features

            # Hidden features
            output_features = made.context_layer.out_features

            print(f"Number of input features: {input_features}")
            print(f"Number of hidden features: {output_features}")

            # Compute posteriors
            print('\n--- Computing posteriors')
            posterior = [model[i].build_posterior(density_estimator[i]) for i in range(len(density_estimator))]

            # Select some random samples
            print(f'\n--- Drawing {n_samples} samples from the held-out dataset')
            if method == 'catch22':
                idx = np.random.choice(theta_EI.shape[0], n_samples, replace=False)
            n_post_samples = 5000
            print(f'--- Sampling {n_post_samples} times the posteriors\n\n')

            # Draw samples from the posteriors
            for xx, sample in enumerate(idx):
                print(f'\r--- Sample {xx + 1}/{len(idx)}, ID: {sample}', end='', flush=True)

                # Features (observation)
                feat = X[sample]
                # Transform the features
                feat = scaler.transform(feat.reshape(1, -1))
                # Torch tensor
                x_o = torch.from_numpy(np.array(feat, dtype=np.float32))

                # Check if the observation is valid
                if torch.isnan(x_o).any():
                    print(f'\n--- Invalid observation: {x_o}')
                    continue

                # Posterior samples
                if use_held_out_data:
                    posterior_samples = [post.sample((n_post_samples,), x=x_o, show_progress_bars=False) for post in
                                         posterior]
                else:
                    fold = int(10. * sample / theta_EI.shape[0])
                    posterior_samples = [posterior[fold].sample((n_post_samples,), x=x_o, show_progress_bars=False)]

                # Calculate E/I
                new_post_samples = [np.zeros((posterior_samples[0].shape[0],
                                              4)) for _ in range(len(posterior_samples))]
                for ii in range(len(posterior_samples)):
                    new_post_samples[ii][:, 0] = (posterior_samples[ii][:, 0] / posterior_samples[ii][:, 2]) /\
                                                 (posterior_samples[ii][:, 1] / posterior_samples[ii][:, 3])
                    new_post_samples[ii][:, 1] = posterior_samples[ii][:, 4]
                    new_post_samples[ii][:, 2] = posterior_samples[ii][:, 5]
                    new_post_samples[ii][:, 3] = posterior_samples[ii][:, 6]

                # Average the sorted posterior samples across folds
                avg_post_samples = np.zeros((new_post_samples[0].shape))
                for ii in range(len(new_post_samples)):
                    avg_post_samples += np.sort(new_post_samples[ii], axis=0)
                avg_post_samples /= len(new_post_samples)

                # Store theta and posterior samples
                all_post_samples[method].append(avg_post_samples)
                all_theta[method].append(theta_EI[sample, :])

                # Diagnostic metrics
                z = np.zeros(4)
                s = np.zeros(4)
                e = np.zeros(4)
                p = np.zeros(4)
                try:
                    for param in range(4):
                        # Remove low-probability samples to prevent distributions with long tails
                        hist, bin_edges = np.histogram(avg_post_samples[:,param],
                                                       bins=np.linspace(lims[param][0], lims[param][1], 1000),
                                                       density=True)
                        hist = gaussian_filter1d(hist, sigma=5)
                        hist /= np.max(hist)
                        iii = bin_edges[np.where(hist > 0.5)[0]]
                        pos = np.where((avg_post_samples[:,param] > iii[0]) & (avg_post_samples[:,param] <= iii[-1]))[0]
                        param_post_samples = avg_post_samples[pos,param]

                        # Calculate z-score
                        z[param] = np.abs( (np.mean(param_post_samples) - theta_EI[sample, param]) /\
                                            np.std(param_post_samples) )

                        # Calculate shrinkage
                        s[param] = 1. - np.var(param_post_samples) / var_theta[param]

                        # # Debug
                        # print(f'--- Parameter {param}:')
                        # print(f'z-score: {z[param]}, shrinkage: {s[param]}')
                        # plt.plot(bin_edges[1:], hist, label='original',alpha=0.5, color='blue', linewidth=1.5)
                        # hist, bin_edges = np.histogram(param_post_samples,
                        #                                bins=np.linspace(lims[param][0], lims[param][1], 1000),
                        #                                density=True)
                        # hist = gaussian_filter1d(hist, sigma=5)
                        # hist /= np.max(hist)
                        # plt.plot(bin_edges[1:], hist, label='filtered',alpha=0.5, color='red', linewidth=1.5)
                        # plt.legend()
                        # plt.show()

                        # Calculate absolute error
                        e[param] = np.abs( np.mean(param_post_samples) - theta_EI[sample, param] )

                        # Calculate PRE for the smallest 25% of the differences between the posterior samples and the
                        # ground truth
                        diff = np.abs(param_post_samples - theta_EI[sample, param])
                        p[param] = np.mean(np.sort(diff)[:int(0.25*len(diff))])
                except:
                    pass

                z_score[method].append(z)
                shrinkage[method].append(s)
                abs_error[method].append(e)
                PRE[method].append(p)

            # Convert to numpy
            all_theta[method] = np.array(all_theta[method])
            z_score[method] = np.array(z_score[method])
            shrinkage[method] = np.array(shrinkage[method])
            abs_error[method] = np.array(abs_error[method])
            PRE[method] = np.array(PRE[method])

        except:
            print(f'\n--- Error loading SBI models for method {method}')
            continue

    # Check if the results folder exists
    if not os.path.exists(result_folder):
        # Try to create folder to save results
        try:
            os.makedirs(result_folder)
        except:
            RuntimeError(f'Could not create folder {result_folder} to save results.')
            sys.exit(1)

    # Save the results
    print('Saving results to file')
    with open(os.path.join(result_folder,'all_post_samples.pkl'), 'wb') as file:
        pickle.dump(all_post_samples, file)
    with open(os.path.join(result_folder,'all_theta.pkl'), 'wb') as file:
        pickle.dump(all_theta, file)
    with open(os.path.join(result_folder,'z_score.pkl'), 'wb') as file:
        pickle.dump(z_score, file)
    with open(os.path.join(result_folder,'shrinkage.pkl'), 'wb') as file:
        pickle.dump(shrinkage, file)
    with open(os.path.join(result_folder,'abs_error.pkl'), 'wb') as file:
        pickle.dump(abs_error, file)
    with open(os.path.join(result_folder,'PRE.pkl'), 'wb') as file:
        pickle.dump(PRE, file)
else:
    print('Loading results from file')
    # Load the results
    with open(os.path.join(result_folder,'all_post_samples.pkl'), 'rb') as file:
        all_post_samples = pickle.load(file)
    with open(os.path.join(result_folder,'all_theta.pkl'), 'rb') as file:
        all_theta = pickle.load(file)
    with open(os.path.join(result_folder,'z_score.pkl'), 'rb') as file:
        z_score = pickle.load(file)
    with open(os.path.join(result_folder,'shrinkage.pkl'), 'rb') as file:
        shrinkage = pickle.load(file)
    with open(os.path.join(result_folder,'abs_error.pkl'), 'rb') as file:
        abs_error = pickle.load(file)
    with open(os.path.join(result_folder,'PRE.pkl'), 'rb') as file:
        PRE = pickle.load(file)


# Pick 2 posteriors to plot
method = 'catch22'

pos = np.argsort(np.sum(z_score[method] + (1 - shrinkage[method]),axis = 1))
s0 = pos[0]
s1 = pos[1]

# # Find a different sample
# for i in np.arange(1,len(pos)):
#     sel = True
#     for param in range(4):
#         diff = np.abs((all_theta[method][pos[i]][param] - all_theta[method][s0][param]) /\
#                       np.max(all_theta[method][:,param]))
#         if  diff < 0.1:
#             sel = False
#     if sel:
#         s1 = pos[i]
#         break

# np.random.seed(int(time.time()))
# s0 = np.random.randint(0, len(all_theta[method]))
# s1 = np.random.randint(0, len(all_theta[method]))

print(f'\n--- Posteriors to plot: {s0} and {s1}')

all_theta_plot = [all_theta[method][s0], all_theta[method][s1]]
all_posterior_plot = [all_post_samples[method][s0], all_post_samples[method][s1]]

# Plots
# Create the figures and set their properties
fig1 = plt.figure(figsize=(7.5, 6), dpi=150)
plt.rcParams.update({'font.size': 10, 'font.family': 'Arial'})
plt.rc('xtick', labelsize=8)
plt.rc('ytick', labelsize=8)

# Pairplot
lims = [[-2, 5], [-1, 3.], [-1, 10], [-5, 70]]
colors = ['blue', 'green']
for row in range(4):
    for col in np.arange(row,4):
        ax = fig1.add_axes([0.04 + col * 0.14, 0.84 - row * 0.14, 0.1, 0.1])

        try:
            # Diagonal: 1D histogram
            if row == col:
                for sample in range(2):
                    hist, bin_edges = np.histogram(all_posterior_plot[sample][:, row],
                                                   bins= np.linspace(lims[row][0], lims[row][1], 50),
                                                   density=True)
                    # 1D smoothing
                    hist = gaussian_filter1d(hist, sigma=1)
                    ax.plot(bin_edges[:-1], hist/np.max(hist),color = colors[sample], alpha = 0.5, linewidth = 2.5)
                    ax.set_xlim(lims[row])

                    # Ground truth values
                    ax.plot([all_theta_plot[sample][row],all_theta_plot[sample][row]], [0,1], color = colors[sample],
                            linewidth = 0.5, linestyle = '--')

                # x-labels
                if row == 0:
                    ax.set_xlabel(r'$E/I$')
                elif row == 1:
                    ax.set_xlabel(r'$\tau_{syn}^{exc}$')
                elif row == 2:
                    ax.set_xlabel(r'$\tau_{syn}^{inh}$')
                else:
                    ax.set_xlabel(r'$J_{syn}^{ext}$')

                # ticks
                ax.set_yticks([])


            # Upper triangle: 2D histogram
            elif row < col:
                for sample in range(2):
                    hist, x_edges, y_edges = np.histogram2d(all_posterior_plot[sample][:, col],
                                                            all_posterior_plot[sample][:, row],
                                                            bins=(np.linspace(lims[col][0], lims[col][1], 25),
                                                                  np.linspace(lims[row][0], lims[row][1], 25)),
                                                            density=True)

                    # Smoothing
                    hist = gaussian_filter1d(hist, sigma=1, axis=0)
                    hist = gaussian_filter1d(hist, sigma=1, axis=1)

                    # Create a custom colormap for this sample
                    cmap = create_white_to_color_cmap(colors[sample])

                    # Plot with transparency
                    ax.pcolormesh(x_edges, y_edges, hist.T, shading='auto', cmap=cmap, vmin=0, vmax=np.max(hist))

                    # Ground truth values
                    ax.scatter([all_theta_plot[sample][col]], [all_theta_plot[sample][row]], s = 5.,
                               c = colors[sample])

                # Set axes limits
                ax.set_xlim(lims[col])
                ax.set_ylim(lims[row])

        except:
            pass

# Legend
ax = fig1.add_axes([0.04, 0.35, 0.2, 0.2])
ax.axis('off')
ax.plot([], [], color='blue', label='sample 1')
ax.plot([], [], color='green', label='sample 2')
# Add ground truth values
ax.plot([], [], color='black', linestyle='--', label='ground truth ('+r'$\theta_0$)')
ax.scatter([], [], color='black', s = 1., label='ground truth ('+r'$\theta_0$)')
ax.legend(loc='upper left', fontsize=8)

# Z-scores versus posterior shrinkage
labels = [r'$E/I$',r'$\tau_{syn}^{exc}$',r'$\tau_{syn}^{inh}$',r'$J_{syn}^{ext}$']
for row in range(3):
    for col in range(2):
        ax = fig1.add_axes([0.66 + col * 0.18, 0.8 - row * 0.21, 0.12, 0.13])

        try:
            method = all_methods[row * 2 + col]

            x = np.linspace(0., 1.05, 8)
            y = np.linspace(0., 10.05, 8)
            hist, x_edges, y_edges = np.histogram2d(shrinkage[method].flatten(),
                                                    z_score[method].flatten(),
                                                    bins=(x, y), density=True)

            # Low-pass filtering
            hist = gaussian_filter1d(hist, sigma=1, axis=0)
            hist = gaussian_filter1d(hist, sigma=1, axis=1)

            mesh = ax.pcolormesh(x_edges, y_edges, hist.T, shading='auto', cmap='Reds')

            # Add a colorbar
            plt.colorbar(mesh, ax=ax)

        except:
            pass

        # x/y-labels
        if row == 2:
            ax.set_xlabel('shrinkage')
        if col == 0:
            ax.set_ylabel('z-score')

        # titles
        if (row * 2 + col) == 0:
            ax.set_title(r'$catch22$', fontsize = 8)
        elif (row * 2 + col) == 1:
            ax.set_title(r'$ch22 + $' + ' ' + r'slp', fontsize = 8)
        elif (row * 2 + col) == 2:
            ax.set_title(r'$dfa$', fontsize = 8)
        elif (row * 2 + col) == 3:
            ax.set_title(r'$rs\ range$', fontsize = 8)
        elif (row * 2 + col) == 4:
            ax.set_title(r'$high\ fluct.$', fontsize = 8)
        elif (row * 2 + col) == 5:
            ax.set_title(r'$1/f$' + ' ' + r'$slope$', fontsize = 8)

# Histograms of errors
colors = ['#FFC0CB', '#FF69B4', '#00FF00', '#32CD32', '#228B22', '#006400']
cmap = LinearSegmentedColormap.from_list('custom_cmap', colors, N=6)

for col in range(4):
    ax = fig1.add_axes([0.08 + col * 0.24, 0.07, 0.18, 0.18])
    all_hist = []

    # Define bins to compute histograms
    if col == 0:
        bins = np.linspace(0, 2., 15)
    elif col == 1:
        bins = np.linspace(0, 1.5, 15)
    elif col == 2:
        bins = np.linspace(0, 6., 15)
    else:
        bins = np.linspace(0, 40., 15)

    for ii,method in enumerate(all_methods):
        try:
            # titles
            if ii == 0:
                label = r'$catch22$'
            elif ii == 1:
                label = r'$ch22 + $'+' '+r'slp'
            elif ii == 2:
                label = r'$dfa$'
            elif ii == 3:
                label = r'$rs\ range$'
            elif ii == 4:
                label = r'$high\ fluct.$'
            elif ii == 5:
                label = r'$1/f$'+' '+r'$slope$'

            # Compute histogram
            hist, bin_edges = np.histogram(abs_error[method][:,col], bins=bins, density=True)

            # Smooth the histogram using a Gaussian filter
            smoothed_hist = gaussian_filter1d(hist, sigma=1)
            all_hist.append(smoothed_hist)
            ax.plot(bin_edges[:-1], smoothed_hist, label=label, color=colors[ii], alpha=0.4, linewidth = 1.5)

        except:
            continue

    # Compute pairwise Hellinger distances
    HD_ = []
    for x in range(6):
        for y in range(x+1, 6):
            HD_.append(hellinger_distance(all_hist[x], all_hist[y]))
            # print('Hellinger distance between', all_methods[x], 'and', all_methods[y], ':', HD_[-1])

    # Show distances
    # Between catch22 and catch22 + slope
    ax.text(0.2 if col > 0 else 0.26, 0.9 if col == 0 else 0.25, r'$D_{H,1}=$'+' %.2f' % HD_[0], fontsize=6, ha='center',
            transform=ax.transAxes, color = colors[1])
    # Between all the remaining methods
    ax.text(0.2, 0.15, r'$D_{H,2}=$'+' %.2f' % np.mean(HD_[9:]), fontsize=6, ha='center', transform=ax.transAxes,
            color = colors[5])
    # Between catch22 and all the remaining methods
    ax.text(0.2, 0.05, r'$D_{H,3}=$'+' %.2f' % np.mean(HD_[1:5]), fontsize=6, ha='center', transform=ax.transAxes,
            color = 'k')

    # legend
    if col == 0:
        ax.legend(loc='upper right', fontsize=6, handletextpad=0.2, borderpad=0.2, labelspacing=0.2)

    # labels
    if col == 0:
        ax.set_ylabel('probability density', fontsize=8)
    ax.set_xlabel('absolute error', fontsize=8)

    # titles
    if col == 0:
        ax.set_title(r'$E/I$', fontsize=10)
    elif col == 1:
        ax.set_title(r'$\tau_{syn}^{exc}$', fontsize=10)
    elif col == 2:
        ax.set_title(r'$\tau_{syn}^{inh}$', fontsize=10)
    else:
        ax.set_title(r'$J_{syn}^{ext}$', fontsize=10)

# Plot letters
ax = fig1.add_axes([0., 0., 1., 1.])
ax.axis('off')
ax.text(0.01, 0.97, 'A', fontsize=12, fontweight='bold')
ax.text(0.59, 0.97, 'B', fontsize=12, fontweight='bold')
ax.text(0.01, 0.28, 'C', fontsize=12, fontweight='bold')

# Save the figure
plt.savefig('SBI_results.png', bbox_inches='tight')
# plt.show()